#    (c)  Copyright 2022 Fungible, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume driver for Fungible Storage Cluster"""

import json
import os
import time

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils

from cinder.common import constants as cinderconstants
from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder.objects import fields
from cinder import utils
from cinder.volume import driver
from cinder.volume.drivers.fungible import constants
from cinder.volume.drivers.fungible import rest_client as rest_api
from cinder.volume.drivers.fungible import swagger_api_client as swagger_client
from cinder.volume.drivers.san import san
from cinder.volume import volume_types
from cinder.volume import volume_utils

LOG = log.getLogger(__name__)

fungible_opts = [
    cfg.PortOpt('nvme_connect_port',
                default=4420,
                help='The port number to be used'
                     ' when doing nvme connect from host'),
    cfg.BoolOpt('api_enable_ssl',
                default=True,
                help='Specify whether to use SSL'
                     ' or not when accessing the composer APIs'),
    cfg.IntOpt('iops_for_image_migration',
               default=250000,
               help='Maximum read IOPS that volume can get'
                    ' when reading data from the volume during'
                    ' host assisted migration'),
    cfg.IntOpt('fsc_clone_volume_timeout',
               default=1800,
               help='Create clone volume timeout in seconds')
]
CONF = cfg.CONF
CONF.register_opts(fungible_opts)


@interface.volumedriver
class FungibleDriver(driver.BaseVD):
    """Fungible Storage driver

    Fungible driver is a volume driver for Fungible Storage.

    Version history:
        1.0.0 - First source driver version
    """

    VERSION = constants.VERSION
    CI_WIKI_NAME = "Fungible_Storage_CI"

    def __init__(self, *args, **kwargs):
        """Initialize the driver."""
        super(FungibleDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(san.san_opts)
        self.configuration.append_config_values(fungible_opts)
        self.rest_client = None
        self.use_multipath = True

    def do_setup(self, context):
        """Initial setup of driver variables"""
        self.rest_client = rest_api.RestClient(self.configuration)
        self.rest_client.do_setup()

    @staticmethod
    def get_driver_options():
        additional_opts = driver.BaseVD._get_oslo_driver_opts(
            "san_ip", "san_login", "san_password", "san_api_port"
        )
        return fungible_opts + additional_opts

    @staticmethod
    def wait_for_device(device_path):
        time.sleep(1)  # wait for a second
        time_to_wait = 4  # 4 seconds
        time_counter = 0
        while not os.path.exists(device_path):
            time.sleep(1)  # wait for a second
            time_counter += 1
            if time_counter > time_to_wait:
                break

    def check_for_setup_error(self):
        """Verify that requirements are in place to use

        Fungible Storage Backend.
        """

        try:
            # backend call for health check
            fungible_res = self.rest_client.check_for_setup_error()
            if fungible_res["status"]:
                LOG.info(
                    "Backend Storage Api Status is %(message)s",
                    {'message': fungible_res['message']})
            else:
                LOG.error(
                    "Backend api status is : %(status)s",
                    {'status': fungible_res['status']})
                raise exception.VolumeBackendAPIException(
                    data=_(
                        "Backend Storage Api Status is "
                        "%(message)s, Error Message: %(err_msg)s)")
                    % {
                        "message": fungible_res["message"],
                        "err_msg": fungible_res["error_message"]
                    }
                )
        except swagger_client.ApiException as e:
            LOG.error(
                "[check_for_setup_error]Request to BackendApi Failed -> %s",
                e.body
            )
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to get backend api status, "
                    "error message: %(error)s." %
                    {'error': error['error_message']}
                )
            )
        except Exception as e:
            LOG.error("[check_for_setup_error]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to get backend api response: %(args)s" %
                    {
                        'args': e.args
                    }
                )
            )

    @staticmethod
    def _get_volume_type_extra_specs(self, volume):
        """Get the Volume type using volume_type_id

        :param: volume object
        :return: volume type & extra specs
        """

        specs = {}
        vol_type = ""
        ctxt = context.get_admin_context()
        type_id = volume["volume_type_id"]
        if type_id:
            LOG.debug("[_get_volume_type_extra_specs]type_id=%s", type_id)
            # get volume type name by volume type id
            volume_type = volume_types.get_volume_type(ctxt, type_id)
            LOG.debug("[_get_volume_type_extra_specs]volume_type=%s",
                      volume_type)
            specs = volume_type.get("extra_specs")
            if constants.FSC_VOL_TYPE in specs:
                vol_type = volume_type.get(
                    "extra_specs").get(constants.FSC_VOL_TYPE)
            else:
                error_msg = (
                    "Key %(type)s was not found in extraspecs" %
                    {
                        'type': constants.FSC_VOL_TYPE
                    }
                )
                LOG.error("[create_volume]Error occurred: %s", error_msg)
                raise exception.VolumeBackendAPIException(
                    data=_(
                        "Failed to create volume %(display_name)s: "
                        "%(error)s." %
                        {'error': error_msg,
                            'display_name': volume.display_name}
                    )
                )

            for key, value in specs.items():
                specs[key] = value
        return specs, vol_type

    def _get_dpu_enabled_host_list(self, ports):
        host_uuid_list = list(
            map(lambda port: port["host_uuid"], ports.values()))
        hosts = self.rest_client.get_hosts_subset(host_uuid_list)
        hosts_fac_enabled = {host["host_uuid"]:
                             host["fac_enabled"] for host in hosts}
        return hosts_fac_enabled

    def create_volume(self, volume):
        """Create volume on Fungible storage backend.

        :param volume: volume to be created
        :return: volume model updates
        """
        fungible_specs = {}
        volume_type = ""
        if "volume_type_id" in volume:
            fungible_specs, volume_type = self._get_volume_type_extra_specs(
                self, volume
            )

        # request fungible to create a volume
        try:
            fungible_res = self.rest_client.create_volume(
                volume, fungible_specs, volume_type
            )
            provider_id = fungible_res["data"]["uuid"]
            # preparing model updates dict to return
            model_updates = {"provider_id": provider_id,
                             "size": volume["size"]}
            LOG.info(
                "Volume created successfully %s. "
                "Volume size: %s. ", volume['id'], volume["size"]
            )
            return model_updates
        except swagger_client.ApiException as e:
            LOG.error(
                "[create_volume]Request to BackendApi Failed -> %s", e.body)
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create volume %(display_name)s: "
                    "%(error)s." %
                    {'error': error['error_message'],
                     'display_name': volume['display_name']}
                )
            )
        except Exception as e:
            LOG.error("[create_volume]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create volume %(name)s: %(args)s" %
                    {
                        'name': volume['display_name'],
                        'args': e.args
                    }
                )
            )

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create Volume on Fungible storage backend

        Args:
            volume: volume to be created
            snapshot: source snapshot from which the volume to be created

        Returns:: volume model updates
        """
        volume_type = ""
        fungible_specs = {}
        if "volume_type_id" in volume:
            fungible_specs, volume_type = self._get_volume_type_extra_specs(
                self, volume
            )

        # request fungible to create a volume
        try:
            fungible_res = self.rest_client.create_volume(
                volume, fungible_specs, volume_type, snapshot
            )
            provider_id = fungible_res["data"]["uuid"]
            # preparing model updates dict to return
            model_updates = {"provider_id": provider_id,
                             "size": volume["size"]}
            LOG.info(
                "Volume created from snapshot successfully with volume "
                "ID: %s. Volume size: %s. ",
                volume['id'], volume['size']
            )
            return model_updates
        except swagger_client.ApiException as e:
            LOG.error(
                "[create_volume_from_snapshot]Request to BackendApi "
                "Failed -> %s", e.body
            )
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create volume from snapshot with volume "
                    "ID: %(name)s: %(error)s." %
                    {'name': volume['display_name'],
                     'error': error['error_message']}
                )
            )
        except Exception as e:
            LOG.error("[create_volume_from_snapshot]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create volume %(name)s: %(args)s" %
                    {
                        'name': volume['display_name'],
                        'args': e.args
                    }
                )
            )

    def delete_volume(self, volume):
        """Delete the available volume

        :param volume: volume to be deleted
        :return: none
        """
        LOG.info("Request to delete volume : %s.", volume['id'])
        if "provider_id" in volume:
            if volume["provider_id"]:
                # request fungible to delete volume
                try:
                    del_res = self.rest_client.delete_volume(
                        volume["provider_id"])
                    LOG.info("Volume delete : %s.", del_res['message'])
                except swagger_client.ApiException as e:
                    LOG.error(
                        "[delete_volume]Request to BackendApi Failed -> %s",
                        e.body
                    )
                    error = json.loads(e.body)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to delete volume "
                            "{volume['display_name']}: "
                            "%(error)s." %
                            {'error': error['error_message']}
                        )
                    )
                except Exception as e:
                    LOG.error("[delete_volume]Error occurred: %s", e)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to delete volume %(name)s: %(args)s" %
                            {
                                'name': volume['display_name'],
                                'args': e.args
                            }
                        )
                    )
            else:
                LOG.info("Volume backend UUID not found in volume details.")
        else:
            raise exception.VolumeBackendAPIException(
                data=_("Failed to delete volume: %s." % volume["id"])
            )

    def create_cloned_volume(self, volume, src_vref):
        """Create volume from volume

        :param volume: volume to be created
        :param src_vref: source volume
        :return: volume model updates
        Logic:
            1. create new volume.
            2. add copy volume task.
            3. in loop check for task status
            4. delete volume copy task
        """

        snapshot_id = None
        try:
            src_volume_uuid = src_vref["provider_id"]

            # create a snapshot to copy the data from
            fungible_res = self.rest_client.create_snapshot(
                src_volume_uuid, src_volume_uuid
            )
            snapshot_id = fungible_res["data"]["uuid"]

            # create new volume.
            new_volume = self.create_volume(volume)
            new_volume_uuid = new_volume.get("provider_id")
            LOG.info(
                "[clone_volume] new volume is created."
                " volume uuid: %s", new_volume_uuid
            )
            # prepare response to return
            model_updates = {"provider_id": new_volume_uuid,
                             "size": volume["size"]}
            # add task to copy volume
            add_task_response = self.rest_client.copy_volume(
                new_volume_uuid, snapshot_id
            )

            # check task status in loop
            task_uuid = add_task_response["data"]["task_uuid"]
            LOG.info(
                "[clone_volume] Copy volume task is added. task_uuid: %s",
                task_uuid
            )
            status = "RUNNING"
            error_message = ""
            sleep_for_seconds = 1
            while status == "RUNNING":
                # Wait before checking for the task status
                # This is done to reduce number of api calls to backend
                # Wait time is increased exponentially to a maximum of 8 secs
                time.sleep(sleep_for_seconds)
                if sleep_for_seconds < 8:
                    sleep_for_seconds = sleep_for_seconds * 2
                task_response = self.rest_client.get_volume_copy_task(
                    task_uuid)
                status = task_response["data"]["task_state"]
                error_message = task_response.get("error_message")

            LOG.info(
                "[clone_volume] Copy volume task with task_uuid:"
                " %s is complete. status: %s", task_uuid, status
            )

            # delete the snapshot created for data copy
            if snapshot_id:
                fungible_res = self.rest_client.delete_snapshot(
                    snapshot_id
                )
                snapshot_id = None
                LOG.info(
                    "Snapshot deleted successfully: %s.",
                    fungible_res['message']
                )

            if status == "FAILED":
                # Delete the new volume created since the data copy failed
                del_res = self.rest_client.delete_volume(new_volume_uuid)
                LOG.info("Volume delete : %s.", del_res['message'])
                raise exception.VolumeBackendAPIException(
                    data=_(
                        "Failed to create new volume %(new_volume_uuid)s: "
                        "from source volume %(src_volume_uuid)s %(error)s." %
                        {
                            'new_volume_uuid': new_volume_uuid,
                            'src_volume_uuid': src_volume_uuid,
                            'error': error_message
                        }
                    )
                )

            try:
                self.rest_client.delete_volume_copy_task(task_uuid)
            except swagger_client.ApiException as e:
                # Just log warning as volume copy is already completed.
                LOG.warning(
                    "[clone_volume] request to delete task %s"
                    " to BackendApi Failed "
                    "-> %s", task_uuid, e.body
                )
        except swagger_client.ApiException as e:
            LOG.error("[clone_volume] request to BackendApi Failed. %s",
                      e.body)
            error = json.loads(e.body)
            # delete the snapshot created for data copy
            if snapshot_id:
                fungible_res = self.rest_client.delete_snapshot(
                    snapshot_id
                )
                snapshot_id = None
                LOG.info(
                    "Snapshot deleted successfully: %s.",
                    fungible_res['message']
                )
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create new volume %(new_volume_uuid)s: "
                    "from source volume %(src_volume_uuid)s %(error)s." %
                    {
                        'new_volume_uuid': new_volume_uuid,
                        'src_volume_uuid': src_volume_uuid,
                        'error': error['error_message']
                    }
                )
            )
        except Exception as e:
            # delete the snapshot created for data copy
            if snapshot_id:
                fungible_res = self.rest_client.delete_snapshot(
                    snapshot_id
                )
                snapshot_id = None
                LOG.info(
                    "Snapshot deleted successfully: %s.",
                    fungible_res['message']
                )
            LOG.error("[create_clone_volume]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create volume %(name)s: %(args)s" %
                    {
                        'name': volume['display_name'],
                        'args': e.args
                    }
                )
            )

        return model_updates

    def ensure_export(self, context, volume):
        pass

    def create_export(self, context, volume, connector):
        pass

    def remove_export(self, context, volume):
        pass

    def initialize_connection(self, volume, connector):
        """Initialize connection and return connection info.

        :param volume: the volume object
        :param connector: the connector object
        :return: connection info dict
        """
        # check for nqn in connector
        host_nqn = connector.get("nqn")
        if not host_nqn:
            host_name = connector.get("host")
            if host_name:
                host_nqn = "nqn.2015-09.com.host:" + host_name
        if not host_nqn:
            raise exception.VolumeBackendAPIException(
                data=_("initialize_connection error: no host nqn available!")
            )

        provider_id = volume.get("provider_id")
        LOG.info("initialize_connection - provider_id=%s", provider_id)

        if not provider_id:
            raise exception.VolumeBackendAPIException(
                data=_("initialize_connection error: no uuid available!")
            )
        try:
            img_mig_iops = False
            # high iops set to true when volume is uploading to image
            # or downloading from image
            if constants.FSC_IOPS_IMG_MIG in connector:
                img_mig_iops = connector.get(constants.FSC_IOPS_IMG_MIG)

            # high iops set to true when volume is migrating
            mig_status = [
                fields.VolumeMigrationStatus.SUCCESS,
            ]
            if volume.get("migration_status") is not None:
                if volume.get("migration_status") not in mig_status:
                    img_mig_iops = True

            # get host_uuid from the host_nqn
            LOG.info("initialize_connection - host_nqn=%s", host_nqn)
            host_uuid = self.rest_client.get_host_uuid_from_host_nqn(host_nqn)
            # create host if it does not exists
            if host_uuid is None:
                host_create_response = self.rest_client.create_host(host_nqn)
                host_uuid = host_create_response["data"]["uuid"]

            LOG.info("initialize_connection - host_uuid=%s", host_uuid)
            host = self.rest_client.get_host_details(host_uuid)

            # request composer to attach volume
            self.rest_client.attach_volume(
                uuid=provider_id,
                host_uuid=host_uuid,
                fac_enabled=host["fac_enabled"],
                iops=img_mig_iops,
            )

            if host["fac_enabled"] is False:
                volume_details = self.rest_client.get_volume_detail(
                    uuid=provider_id)

                target_nqn = volume_details.get("data").get("subsys_nqn")
                get_config_value = self.configuration.safe_get
                port = get_config_value("nvme_connect_port")

                topology_response = self.rest_client.get_topology()
                LOG.info(
                    "initialize_connection - topology_response=%s",
                    topology_response
                )

                str_portals = []
                # find primary dpu ip
                primary_dpu = volume_details.get("data").get("dpu")
                LOG.info("initialize_connection - primary_dpu=%s",
                         primary_dpu)
                if primary_dpu:
                    if topology_response["status"] is True:
                        topology_data = topology_response.get("data")
                        for device in topology_data.values():
                            for dpu in device["dpus"]:
                                if dpu["uuid"] == primary_dpu:
                                    portal_ip = str(dpu["dataplane_ip"])
                                    portal_port = str(port)
                                    portal_transport = "tcp"
                                    str_portals.append(
                                        (
                                            portal_ip,
                                            portal_port,
                                            portal_transport
                                        )
                                    )
                # find secondary dpu ip
                secondary_dpu = volume_details.get("data").get("secy_dpu")
                LOG.info(
                    "initialize_connection - secondary_dpu=%s", secondary_dpu)
                if secondary_dpu:
                    if topology_response["status"] is True:
                        topology_data = topology_response.get("data")
                        for device in topology_data.values():
                            for dpu in device["dpus"]:
                                if dpu["uuid"] == secondary_dpu:
                                    portal_ip = str(dpu["dataplane_ip"])
                                    portal_port = str(port)
                                    portal_transport = "tcp"
                                    str_portals.append(
                                        (
                                            portal_ip,
                                            portal_port,
                                            portal_transport
                                        )
                                    )

                # preparing connection info dict to return
                vol_nguid = provider_id.replace("-", "")
                data = {
                    "vol_uuid": provider_id,
                    "target_nqn": str(target_nqn),
                    "host_nqn": host_nqn,
                    "portals": str_portals,
                    "volume_nguid": vol_nguid,
                }
                conn_info = {"driver_volume_type": "nvmeof", "data": data}
                LOG.info("initialize_connection - conn_info=%s", conn_info)
            else:
                raise exception.VolumeBackendAPIException(
                    data=_("FAC enabled hosts are not supported")
                )

            return conn_info
        except swagger_client.ApiException as e:
            LOG.error(
                "[initialize_connection]Request to BackendApi Failed -> %s",
                e.body
            )
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to attach the volume %(name)s: %(error)s." %
                    {
                        'name': volume.get('display_name'),
                        'error': error['error_message']
                    }
                )
            )
        except Exception as e:
            LOG.error("[initialize_connection]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to attach volume %(name)s: %(args)s" %
                    {
                        'name': volume.get('display_name'),
                        'args': e.args
                    }
                )
            )

    def terminate_connection(self, volume, connector, **kwargs):
        """Terminate connection for detaching the port from volume.

        :param volume: the volume object
        :param connector: the connector object
        """
        provider_id = volume.get("provider_id")
        LOG.info("terminate_connection - provider_id=%s", provider_id)
        if not provider_id:
            raise exception.VolumeBackendAPIException(
                data=_("terminate_connection error: no provider_id available.")
            )

        try:
            volume_details = self.rest_client.get_volume_detail(
                uuid=provider_id)
            LOG.info("terminate_connection - volume_details=%s",
                     volume_details)

            if connector is None:
                # None connector means force-detach
                # Remove all ports from backend
                ports = volume_details["data"]["ports"]

                if ports:
                    # Get the host details for each attachment
                    hosts_fac_enabled = self._get_dpu_enabled_host_list(ports)
                    # request composer to detach volume
                    for port_id in ports.keys():
                        if (
                            ports.get(port_id)["transport"] == "PCI"
                            or not hosts_fac_enabled[
                                ports.get(port_id)["host_uuid"]
                            ]
                        ):
                            self.rest_client.detach_volume(port_id)

                LOG.info("Removed all the ports from storage backend.")
                return

            host_nqn = connector.get("nqn")
            if not host_nqn:
                host_name = connector.get("host")
                if host_name:
                    host_nqn = "nqn.2015-09.com.host:" + host_name
            if not host_nqn:
                raise exception.VolumeBackendAPIException(
                    data=_("terminate_connection error: "
                           "no host nqn available.")
                )

            # get host_uuid from the host_nqn
            LOG.info("terminate_connection - host_nqn=%s", host_nqn)
            host_uuid = self.rest_client.get_host_uuid_from_host_nqn(host_nqn)
            LOG.info("terminate_connection - host_uuid=%s", host_uuid)

            ports = volume_details["data"]["ports"]
            if host_uuid and ports:
                port_ids = [
                    port
                    for port in ports.keys()
                    if ports.get(port)["host_uuid"] == host_uuid
                ]
                # request fungible to detach volume
                if port_ids:
                    # Get the host details for each attachment
                    hosts_fac_enabled = self._get_dpu_enabled_host_list(ports)
                    # request composer to detach volume
                    for port_id in port_ids:
                        if (
                            ports.get(port_id)["transport"] == "PCI"
                            or not hosts_fac_enabled[
                                ports.get(port_id)["host_uuid"]
                            ]
                        ):
                            self.rest_client.detach_volume(port_id)
                    LOG.info(
                        "Volume detached successfully. \
                            provider_id=%s", provider_id
                    )
                else:
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "terminate_connection error: "
                            "required port is not available for detach."
                        )
                    )
            else:
                raise exception.VolumeBackendAPIException(
                    data=_(
                        "terminate_connection error: "
                        "Volume not attached to any ports."
                    )
                )
        except swagger_client.ApiException as e:
            LOG.error(
                "[terminate_connection]Request to BackendApi Failed -> %s",
                e.body
            )
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to detach the volume "
                    "%(name)s: %(error)s." %
                    {
                        'name': volume.get('display_name'),
                        'error': error['error_message']
                    }
                )
            )
        except Exception as e:
            LOG.error("[terminate_connection]Error occurred: %s", e)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to detach volume %(name)s: %(args)s" %
                    {
                        'name': volume.get('display_name'),
                        'args': e.args
                    }
                )
            )

    def create_snapshot(self, snapshot):
        """Create volume snapshot on storage backend.

        :param snapshot: volume snapshot to be created
        :return: snapshot model updates
        """
        if "provider_id" in snapshot.volume:
            if snapshot.volume.provider_id:
                try:
                    # request fungible to create snapshot
                    fungible_res = self.rest_client.create_snapshot(
                        snapshot.volume.provider_id, snapshot.id
                    )
                    provider_id = fungible_res["data"]["uuid"]
                    # fungible model updates dict to return
                    model_updates = {
                        "provider_id": provider_id,
                    }
                    LOG.info(
                        "Snapshot created successfully %s. ",
                        snapshot.id)
                    return model_updates
                except swagger_client.ApiException as e:
                    LOG.error(
                        "[create_snapshot]Request to BackendApi Failed -> %s",
                        e.body
                    )
                    error = json.loads(e.body)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to create the snapshot "
                            "%(name)s: %(error)s."
                        ) % {
                            'name': snapshot.display_name,
                            'error': error['error_message']
                        }
                    )
                except Exception as e:
                    LOG.error("[create_snapshot]Error occurred: %s", e)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to create snapshot %(name)s: %(args)s" %
                            {
                                'name': snapshot.display_name,
                                'args': e.args
                            }
                        )
                    )
            else:
                raise exception.VolumeBackendAPIException(
                    data=_(
                        "Failed to create snapshot: volume provider_id "
                        "not found in snapshot's volume details."
                    )
                )
        else:
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to create snapshot, volume provider_id attribute "
                    "not found in snapshot details :%s." % snapshot.id
                )
            )

    def delete_snapshot(self, snapshot):
        """Delete snapshot from storage backend.

        :param snapshot: snapshot to be deleted
        """
        LOG.info("Request to delete snapshot : %s.", snapshot['id'])
        if "provider_id" in snapshot:
            if snapshot["provider_id"]:
                try:
                    # request fungible to delete snapshot
                    fungible_res = self.rest_client.delete_snapshot(
                        snapshot["provider_id"]
                    )
                    LOG.info(
                        "Snapshot deleted successfully: %s.",
                        fungible_res['message']
                    )
                except swagger_client.ApiException as e:
                    LOG.error(
                        "[delete_snapshot]Request to BackendApi Failed -> %s",
                        e.body
                    )
                    error = json.loads(e.body)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to delete the snapshot "
                            "%(name)s: %(error)s." %
                            {
                                'name': snapshot['display_name'],
                                'error': error['error_message']
                            }
                        )
                    )
                except Exception as e:
                    LOG.error("[delete_snapshot]Error occurred: %s", e)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to delete snapshot %(name)s: %(args)s" %
                            {
                                'name': snapshot['display_name'],
                                'args': e.args
                            }
                        )
                    )
            else:
                LOG.info("Snapshot backend UUID not found in snapshot "
                         "details.")
        else:
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to delete snapshot, provider_id attribute "
                    "not found in snapshot details :%s." % snapshot["id"]
                )
            )

    def extend_volume(self, volume, new_size):
        """Extend size of existing fungible volume.

        :param volume: volume to be extended
        :param new_size: volume size after extending
        """
        LOG.info("Request to extend volume : %s.", volume['id'])
        if "provider_id" in volume:
            if volume["provider_id"]:
                try:
                    # request fungible to extend volume
                    self.rest_client.extend_volume(
                        volume["provider_id"], new_size)
                    LOG.info(
                        "Volume %s is resized successfully", volume['id'])
                except swagger_client.ApiException as e:
                    LOG.error(
                        "[extend_volume]Request to BackendApi Failed -> %s",
                        e.body
                    )
                    error = json.loads(e.body)
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to extend the volume "
                            "%(name)s: %(error)s." %
                            {
                                'name': volume.get('display_name'),
                                'error': error['error_message']
                            }
                        )
                    )
                except Exception as e:
                    LOG.error("[extend_volume]Error occurred: {e}")
                    raise exception.VolumeBackendAPIException(
                        data=_(
                            "Failed to extend volume %(name)s: %(args)s" %
                            {
                                'name': volume.get('display_name'),
                                'args': e.args
                            }
                        )
                    )
            else:
                LOG.warning(
                    "Volume backend UUID not found in volume details.")
        else:
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to extend volume, provider_id attribute "
                    "not found in volume details :%s." % volume["id"]
                )
            )

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image."""
        LOG.info(
            "Copy volume %s to image on "
            "image service %s. Image meta: %s.",
            volume['id'], image_service, image_meta
        )

        use_multipath = self.configuration.use_multipath_for_image_xfer
        enforce_multipath = self.configuration.enforce_multipath_for_image_xfer
        if hasattr(utils, "brick_get_connector_properties"):
            properties = utils.brick_get_connector_properties(
                use_multipath, enforce_multipath
            )
        else:
            properties = volume_utils.brick_get_connector_properties(
                use_multipath, enforce_multipath
            )
        # added iops parameter in properties to
        # perform high iops while uploading volume to image
        properties[constants.FSC_IOPS_IMG_MIG] = True
        attach_info, volume = self._attach_volume(context, volume, properties)

        try:
            # Wait until the device path appears
            self.wait_for_device(attach_info["device"]["path"])
            image_utils.upload_volume(
                context,
                image_service,
                image_meta,
                attach_info["device"]["path"],
                compress=True,
            )
            LOG.debug(
                "Copy volume %s to image complete",
                volume['id']
            )
        finally:
            # Since attached volume was not used for writing we can force
            # detach it
            self._detach_volume(
                context, attach_info, volume, properties, force=True,
                ignore_errors=True
            )

    def copy_image_to_volume(self, context, volume, image_service, image_id,
                             disable_sparse=False):
        """Fetch the image from image_service and write it to the volume."""
        LOG.info(
            "Copy image %s from image service %s "
            "to volume %s.", image_id, image_service, volume['id']
        )

        use_multipath = self.configuration.use_multipath_for_image_xfer
        enforce_multipath = self.configuration.enforce_multipath_for_image_xfer
        if hasattr(utils, "brick_get_connector_properties"):
            properties = utils.brick_get_connector_properties(
                use_multipath, enforce_multipath
            )
        else:
            properties = volume_utils.brick_get_connector_properties(
                use_multipath, enforce_multipath
            )
        # added iops parameter in properties to
        # perform high iops while downloading image to volume
        properties[constants.FSC_IOPS_IMG_MIG] = True
        attach_info, volume = self._attach_volume(context, volume, properties)
        try:
            # Wait until the device path appears
            self.wait_for_device(attach_info["device"]["path"])
            image_utils.fetch_to_raw(
                context,
                image_service,
                image_id,
                attach_info["device"]["path"],
                self.configuration.volume_dd_blocksize,
                size=volume["size"],
                disable_sparse=disable_sparse,
            )
            LOG.debug(
                "Copy image %s to volume %s complete",
                image_id, volume['id']
            )
        except exception.ImageTooBig:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "Copying image %(image_id)s to "
                    "volume failed due to insufficient available "
                    "space.",
                    {"image_id": image_id},
                )
        finally:
            self._detach_volume(context, attach_info,
                                volume, properties, force=True)

    def update_migrated_volume(self, ctxt, volume, new_volume,
                               original_volume_status):
        """Update volume name of new fungible volume.

        Original volume is renamed first since fungible does not allow
        multiple volumes to have same name.
        """
        try:
            new_name = volume["id"]
            LOG.info("Rename volume from %s to %s.", new_volume['id'],
                     new_name)
            LOG.info("Update backend volume name to %s", new_name)
            # if new volume provider id is None, # volume will not be renamed.
            if new_volume["provider_id"]:
                # if original provider id is None & volume host doesn't match,
                # original volume will not be renamed
                if volume["provider_id"] and (volume["host"] ==
                                              new_volume["host"]):
                    try:
                        self.rest_client.rename_volume(
                            volume["provider_id"], "migrating_" + new_name
                        )
                    except swagger_client.ApiException as e:
                        LOG.warning(
                            "Failed to rename the original volume %s.",
                            e.body
                        )
                else:
                    LOG.warning(
                        "Original volume backend UUID not found in "
                        "volume details."
                    )
                self.rest_client.rename_volume(
                    new_volume["provider_id"], new_name)
            else:
                LOG.warning(
                    "New volume backend UUID not found in volume details.")
            return {"_name_id": None}
        except swagger_client.ApiException as e:
            LOG.error(
                "[update_migrated_volume]Request to BackendApi Failed -> %s",
                e.body
            )
            error = json.loads(e.body)
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to rename the volume %(name)s:"
                    " %(error)s." %
                    {
                        'name': volume.get('display_name'),
                        'error': error['error_message']
                    }
                )
            )
        except Exception as e:
            LOG.error("[update_migrated_volume]Error occurred: {e}")
            raise exception.VolumeBackendAPIException(
                data=_(
                    "Failed to rename volume %(name)s: %(args)s" %
                    {
                        'name': volume.get('display_name'),
                        'args': e.args
                    }
                )
            )

    def get_volume_stats(self, refresh=False):
        """Get the volume stats"""
        data = {
            "volume_backend_name":
            self.configuration.safe_get("volume_backend_name"),
            "vendor_name": "Fungible Inc.",
            "driver_version": self.VERSION,
            "storage_protocol": cinderconstants.NVMEOF_TCP,
            "total_capacity_gb": "unknown",
            "free_capacity_gb": "unknown",
        }
        return data



# Copyright 2024 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for the image format_inspector.
"""


from cinder.image import format_inspector
import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def get_format_if_safe(path, allow_qcow2_backing_file):
    """Returns a str format name if the format is safe, otherwise None"""
    return _get_format_if_safe(path, allow_qcow2_backing_file)


def _get_format_if_safe(path, allow_qcow2_backing_file):
    """Returns a str format name if the format is safe, otherwise None"""
    inspector = format_inspector.detect_file_format(path)
    format_name = str(inspector)
    safe = inspector.safety_check()
    if not safe and format_name == 'qcow2' and allow_qcow2_backing_file:
        safe = inspector.safety_check_allow_backing_file()
    if safe:
        return format_name



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import timeutils

from cinder import context
from cinder import exception
from cinder.tests.unit.targets import targets_fixture as tf
from cinder import utils
from cinder.volume.targets import nvmeof


class FakeNVMeOFDriver(nvmeof.NVMeOF):
    def __init__(self, *args, **kwargs):
        super(FakeNVMeOFDriver, self).__init__(*args, **kwargs)

    def delete_nvmeof_target(self, target_name):
        pass


@ddt.ddt
class TestNVMeOFDriver(tf.TargetDriverFixture):

    def setUp(self):
        super(TestNVMeOFDriver, self).setUp()

        self.configuration.target_protocol = 'nvmet_rdma'
        self.target = FakeNVMeOFDriver(root_helper=utils.get_root_helper(),
                                       configuration=self.configuration)

        self.target_ip = self.configuration.target_ip_address
        self.target_port = self.configuration.target_port
        self.nvmet_subsystem_name = self.configuration.target_prefix
        self.nvmet_ns_id = self.configuration.nvmet_ns_id
        self.nvmet_port_id = self.configuration.nvmet_port_id
        self.nvme_transport_type = 'rdma'

        self.fake_volume_id = 'c446b9a2-c968-4260-b95f-a18a7b41c004'
        self.testvol_path = (
            '/dev/stack-volumes-lvmdriver-1/volume-%s' % self.fake_volume_id)
        self.fake_project_id = 'ed2c1fd4-5555-1111-aa15-123b93f75cba'
        self.testvol = (
            {'project_id': self.fake_project_id,
             'name': 'testvol',
             'size': 1,
             'id': self.fake_volume_id,
             'volume_type_id': None,
             'provider_location':
                 self.target.get_nvmeof_location(
                     "ngn.%s-%s" % (
                         self.nvmet_subsystem_name,
                         self.fake_volume_id),
                     [self.target_ip],
                     self.target_port,
                     self.nvme_transport_type,
                     self.nvmet_ns_id
                 ),
             'provider_auth': None,
             'provider_geometry': None,
             'created_at': timeutils.utcnow(),
             'host': 'fake_host@lvm#lvm'})

    @mock.patch.object(nvmeof.NVMeOF, '_get_connection_properties_from_vol')
    def test_initialize_connection(self, mock_get_conn):
        mock_connector = {'initiator': 'fake_init'}
        mock_testvol = self.testvol
        expected_return = {
            'driver_volume_type': 'nvmeof',
            'data': mock_get_conn.return_value
        }
        self.assertEqual(expected_return,
                         self.target.initialize_connection(mock_testvol,
                                                           mock_connector))
        mock_get_conn.assert_called_once_with(mock_testvol)

    @mock.patch.object(FakeNVMeOFDriver, 'create_nvmeof_target')
    def test_create_export(self, mock_create_nvme_target):
        ctxt = context.get_admin_context()
        self.target.create_export(ctxt, self.testvol, self.testvol_path)
        mock_create_nvme_target.assert_called_once_with(
            self.fake_volume_id,
            self.configuration.target_prefix,
            [self.target_ip],
            self.target_port,
            self.nvme_transport_type,
            self.nvmet_port_id,
            self.nvmet_ns_id,
            self.testvol_path
        )

    @mock.patch.object(FakeNVMeOFDriver, 'delete_nvmeof_target')
    def test_remove_export(self, mock_delete_nvmeof_target):
        ctxt = context.get_admin_context()
        self.target.remove_export(ctxt, self.testvol)
        mock_delete_nvmeof_target.assert_called_once_with(
            self.testvol
        )

    @mock.patch.object(nvmeof.NVMeOF, '_get_nvme_uuid')
    @mock.patch.object(nvmeof.NVMeOF, '_get_connection_properties')
    def test__get_connection_properties(self, mock_get_conn_props, mock_uuid):
        """Test connection properties from a volume."""
        res = self.target._get_connection_properties_from_vol(self.testvol)
        self.assertEqual(mock_get_conn_props.return_value, res)
        mock_uuid.assert_called_once_with(self.testvol)
        mock_get_conn_props.assert_called_once_with(
            f'ngn.{self.nvmet_subsystem_name}-{self.fake_volume_id}',
            [self.target_ip],
            str(self.target_port),
            self.nvme_transport_type,
            str(self.nvmet_ns_id),
            mock_uuid.return_value)

    @mock.patch.object(nvmeof.NVMeOF, '_get_nvme_uuid')
    @mock.patch.object(nvmeof.NVMeOF, '_get_connection_properties')
    def test__get_connection_properties_multiple_addresses(
            self, mock_get_conn_props, mock_uuid):
        """Test connection properties from a volume with multiple ips."""
        self.testvol['provider_location'] = self.target.get_nvmeof_location(
            f"ngn.{self.nvmet_subsystem_name}-{self.fake_volume_id}",
            [self.target_ip, '127.0.0.1'],
            self.target_port,
            self.nvme_transport_type,
            self.nvmet_ns_id
        )

        res = self.target._get_connection_properties_from_vol(self.testvol)
        self.assertEqual(mock_get_conn_props.return_value, res)
        mock_uuid.assert_called_once_with(self.testvol)
        mock_get_conn_props.assert_called_once_with(
            f'ngn.{self.nvmet_subsystem_name}-{self.fake_volume_id}',
            [self.target_ip, '127.0.0.1'],
            str(self.target_port),
            self.nvme_transport_type,
            str(self.nvmet_ns_id),
            mock_uuid.return_value)

    def test__get_connection_properties_old(self):
        """Test connection properties with the old NVMe-oF format."""
        nqn = f'ngn.{self.nvmet_subsystem_name}-{self.fake_volume_id}'
        expected_return = {
            'target_portal': self.target_ip,
            'target_port': str(self.target_port),
            'nqn': nqn,
            'transport_type': self.nvme_transport_type,
            'ns_id': str(self.nvmet_ns_id)
        }
        res = self.target._get_connection_properties(nqn,
                                                     [self.target_ip],
                                                     str(self.target_port),
                                                     self.nvme_transport_type,
                                                     str(self.nvmet_ns_id),
                                                     mock.sentinel.uuid)
        self.assertEqual(expected_return, res)

    @ddt.data(('rdma', 'RoCEv2'), ('tcp', 'tcp'))
    @ddt.unpack
    def test__get_connection_properties_new(
            self, transport, expected_transport):
        """Test connection properties with the new NVMe-oF format."""
        nqn = f'ngn.{self.nvmet_subsystem_name}-{self.fake_volume_id}'
        self.configuration.nvmeof_conn_info_version = 2

        expected_return = {
            'target_nqn': nqn,
            'vol_uuid': mock.sentinel.uuid,
            'ns_id': str(self.nvmet_ns_id),
            'portals': [(self.target_ip,
                         str(self.target_port),
                         expected_transport)],
        }
        res = self.target._get_connection_properties(nqn,
                                                     [self.target_ip],
                                                     str(self.target_port),
                                                     transport,
                                                     str(self.nvmet_ns_id),
                                                     mock.sentinel.uuid)
        self.assertEqual(expected_return, res)

    @ddt.data({'nqn': 'fake-nqn'},
              {'nqn': 'fake-nqn', 'initiator': 'fake-iqn'})
    def test_validate_connector(self, mock_connector):
        self.assertTrue(self.target.validate_connector(mock_connector))

    @ddt.data({'initiator': 'fake-iqn'}, {})
    def test_validate_connector_not_found(self, mock_connector):
        self.assertRaises(exception.InvalidConnectorException,
                          self.target.validate_connector,
                          mock_connector)

    def test_invalid_target_protocol(self):
        self.configuration.target_protocol = 'iser'
        self.assertRaises(nvmeof.UnsupportedNVMETProtocol,
                          FakeNVMeOFDriver,
                          root_helper=utils.get_root_helper(),
                          configuration=self.configuration)

    def test_invalid_secondary_ips_old_conn_info_combination(self):
        """Secondary IPS are only supported with new connection information."""
        self.configuration.target_secondary_ip_addresses = ['127.0.0.1']
        self.configuration.nvmeof_conn_info_version = 1
        self.assertRaises(exception.InvalidConfigurationValue,
                          FakeNVMeOFDriver,
                          root_helper=utils.get_root_helper(),
                          configuration=self.configuration)

    def test_valid_secondary_ips_old_conn_info_combination(self):
        """Secondary IPS are supported with new connection information."""
        self.configuration.target_secondary_ip_addresses = ['127.0.0.1']
        self.configuration.nvmeof_conn_info_version = 2
        FakeNVMeOFDriver(root_helper=utils.get_root_helper(),
                         configuration=self.configuration)

    def test_are_same_connector(self):
        res = self.target.are_same_connector({'nqn': 'nvme'}, {'nqn': 'nvme'})
        self.assertTrue(res)

    @ddt.data(({}, {}), ({}, {'nqn': 'nvmE'}), ({'nqn': 'nvmeE'}, {}),
              ({'nqn': 'nvme1'}, {'nqn': 'nvme2'}))
    @ddt.unpack
    def test_are_same_connector_different(self, a_conn_props, b_conn_props):
        res = self.target.are_same_connector(a_conn_props, b_conn_props)
        self.assertFalse(bool(res))

    def test_get_nvmeof_location(self):
        """Serialize connection information into location."""
        result = self.target.get_nvmeof_location(
            'ngn.subsys_name-vol_id', ['127.0.0.1'], 4420, 'tcp', 10)

        expected = '127.0.0.1:4420 tcp ngn.subsys_name-vol_id 10'
        self.assertEqual(expected, result)

    def test_get_nvmeof_location_multiple_ips(self):
        """Serialize connection information with multiple ips into location."""
        result = self.target.get_nvmeof_location(
            'ngn.subsys_name-vol_id', ['127.0.0.1', '192.168.1.1'], 4420,
            'tcp', 10)

        expected = '127.0.0.1,192.168.1.1:4420 tcp ngn.subsys_name-vol_id 10'
        self.assertEqual(expected, result)



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

import ddt
import requests.exceptions

from cinder import exception
from cinder.tests.unit import fake_constants
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powerstore import MockResponse
from cinder.volume.drivers.dell_emc.powerstore import (
    exception as powerstore_exception)
from cinder.volume.drivers.dell_emc.powerstore import client


CLIENT_OPTIONS = {
    "rest_ip": "127.0.0.1",
    "rest_username": "fake_user",
    "rest_password": "fake_password",
    "verify_certificate": False,
    "certificate_path": None,
    "rest_api_connect_timeout": 60,
    "rest_api_read_timeout": 60
}

ISCSI_IP_POOL_RESP = [
    {
        "address": "1.2.3.4",
        "ip_port": {
            "target_iqn":
                "iqn.2022-07.com.dell:dellemc-powerstore-fake-iqn-1"
        },
    },
    {
        "address": "5.6.7.8",
        "ip_port": {
            "target_iqn":
                "iqn.2022-07.com.dell:dellemc-powerstore-fake-iqn-1"
        },
    },
]

NVME_IP_POOL_RESP = [
    {
        "address": "11.22.33.44"
    },
    {
        "address": "55.66.77.88"
    }
]

QOS_IO_RULE_PARAMS = {
    "name": "io-rule-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf",
    "type": "Absolute",
    "max_iops": "200",
    "max_bw": "18000",
    "burst_percentage": "50"
}

QOS_POLICY_PARAMS = {
    "name": "qos-policy-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf",
    "io_limit_rule_id": "9beb10ff-a00c-4d88-a7d9-692be2b3073f"
}

QOS_UPDATE_IO_RULE_PARAMS = {
    "type": "Absolute",
    "max_iops": "500",
    "max_bw": "225000",
    "burst_percentage": "89"
}


@ddt.ddt
class TestClient(test.TestCase):

    def setUp(self):
        super(TestClient, self).setUp()

        self.client = client.PowerStoreClient(**CLIENT_OPTIONS)
        self.fake_volume = str(uuid.uuid4())

    @ddt.data(("iSCSI", ISCSI_IP_POOL_RESP),
              ("NVMe", NVME_IP_POOL_RESP))
    @ddt.unpack
    @mock.patch("requests.request")
    def test_get_ip_pool_address(self, protocol, ip_pool, mock_request):
        mock_request.return_value = MockResponse(ip_pool, rc=200)
        response = self.client.get_ip_pool_address(protocol)
        mock_request.assert_called_once()
        self.assertEqual(response, ip_pool)

    @mock.patch("requests.request")
    def test_get_volume_nguid(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "nguid": "nguid.76e02b0999y439958ttf546800ea7fe8"
            },
            rc=200
        )
        self.assertEqual(self.client.get_volume_nguid(self.fake_volume),
                         "76e02b0999y439958ttf546800ea7fe8")

    @mock.patch("requests.request")
    def test_get_array_version(self, mock_request):
        mock_request.return_value = MockResponse(
            content=[
                {
                    "release_version": "3.0.0.0",
                }
            ],
            rc=200
        )
        self.assertEqual(self.client.get_array_version(),
                         "3.0.0.0")

    @mock.patch("requests.request")
    def test_get_qos_policy_id_by_name(self, mock_request):
        mock_request.return_value = MockResponse(
            content=[
                {
                    "id": "d69f7131-4617-4bae-89f8-a540a6bda94b",
                }
            ],
            rc=200
        )
        self.assertEqual(
            self.client.get_qos_policy_id_by_name("qos-"
                                                  "policy-6b6e5489"
                                                  "-4b5b-4468-a1f7-"
                                                  "32cec2ffa3bf"),
            "d69f7131-4617-4bae-89f8-a540a6bda94b")

    @mock.patch("requests.request")
    def test_get_qos_policy_id_by_name_exception(self, mock_request):
        mock_request.return_value = MockResponse(rc=400)
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.client.get_qos_policy_id_by_name,
            "qos-policy-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf")

    @mock.patch("requests.request")
    def test_create_qos_io_rule(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "id": "9beb10ff-a00c-4d88-a7d9-692be2b3073f"
            },
            rc=200
        )
        self.assertEqual(
            self.client.create_qos_io_rule(QOS_IO_RULE_PARAMS),
            "9beb10ff-a00c-4d88-a7d9-692be2b3073f")

    @mock.patch("requests.request")
    def test_create_duplicate_qos_io_rule(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "messages": [
                    {
                        "code": "0xE0A0E0010009",
                        "severity": "Error",
                        "message_l10n": "The rule name "
                        "io-rule-9899a65f-70fe-46c9-8f6c-22625c7e19df "
                        "is already used by another rule. "
                        "It needs to be unique (case-insensitive). "
                        "Please use a different name.",
                        "arguments": [
                            "io-rule-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf"
                        ]
                    }
                ]
            },
            rc=400
        )
        self.assertRaises(
            powerstore_exception.DellPowerStoreQoSIORuleExists,
            self.client.create_qos_io_rule,
            QOS_IO_RULE_PARAMS)

    @mock.patch("requests.request")
    def test_create_duplicate_qos_io_rule_with_unexpected_error(
            self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "messages": [
                    {
                        "code": "0xE0101001000C",
                        "severity": "Error",
                        "message_l10n": "The system encountered unexpected "
                                        "backend errors. "
                                        "Please contact support."
                    }
                ]
            },
            rc=400
        )
        self.assertRaises(
            powerstore_exception.DellPowerStoreQoSIORuleExists,
            self.client.create_qos_io_rule,
            QOS_IO_RULE_PARAMS)

    @mock.patch("requests.request")
    def test_create_qos_policy(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "id": "d69f7131-4617-4bae-89f8-a540a6bda94b",
            },
            rc=200
        )
        self.assertEqual(
            self.client.create_qos_policy(QOS_POLICY_PARAMS),
            "d69f7131-4617-4bae-89f8-a540a6bda94b")

    @mock.patch("requests.request")
    def test_create_duplicate_qos_policy(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "messages": [
                    {
                        "code": "0xE02020010004",
                        "severity": "Error",
                        "message_l10n": "The new policy name qos-policy-"
                                        "6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf "
                                        "is in use. It must be unique "
                                        "regardless of character cases.",
                        "arguments": [
                            "qos-policy-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf"
                        ]
                    }
                ]
            },
            rc=400
        )
        self.assertRaises(
            powerstore_exception.DellPowerStoreQoSPolicyExists,
            self.client.create_qos_policy,
            QOS_POLICY_PARAMS)

    @mock.patch("requests.request")
    def test_update_volume_with_qos_policy(self, mock_request):
        mock_request.return_value = MockResponse(rc=200)
        self.client.update_volume_with_qos_policy(
            fake_constants.VOLUME_ID,
            "qos-policy-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf")
        mock_request.assert_called_once()

    @mock.patch("requests.request")
    def test_update_volume_with_qos_policy_exception(self, mock_request):
        mock_request.return_value = MockResponse(rc=400)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.client.update_volume_with_qos_policy,
                          fake_constants.VOLUME_ID,
                          "qos-policy-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf")

    @mock.patch("requests.request")
    def test_update_qos_io_rule(self, mock_request):
        mock_request.return_value = MockResponse(rc=200)
        self.client.update_qos_io_rule(
            "io-rule-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf",
            QOS_UPDATE_IO_RULE_PARAMS)
        mock_request.assert_called_once()

    @mock.patch("requests.request")
    def test_update_qos_io_rule_exception(self, mock_request):
        mock_request.return_value = MockResponse(rc=400)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.client.update_qos_io_rule,
                          "io-rule-6b6e5489-4b5b-4468-a1f7-32cec2ffa3bf",
                          QOS_UPDATE_IO_RULE_PARAMS)

    @mock.patch("requests.request")
    def test_get_request_timeout_exception(self, mock_request):
        mock_request.return_value = MockResponse(
            rc=501
        )
        error = self.assertRaises(
            exception.VolumeBackendAPIException,
            self.client.get_array_version)
        self.assertEqual('Bad or unexpected response from the '
                         'storage volume backend API: Failed to '
                         'query PowerStore array version.',
                         error.msg)

    @mock.patch("requests.request")
    def test_send_get_request_connect_timeout_exception(self,
                                                        mock_request):
        mock_request.side_effect = requests.exceptions.ConnectTimeout()
        r, resp = self.client._send_request("GET",
                                            "/api/version")
        self.assertEqual(500, r.status_code)

    @mock.patch("requests.request")
    def test_send_get_request_read_timeout_exception(self,
                                                     mock_request):
        mock_request.side_effect = requests.exceptions.ReadTimeout()
        r, resp = self.client._send_request("GET",
                                            "/api/version")
        self.assertEqual(500, r.status_code)

    @mock.patch("requests.request")
    def test_send_post_request_connect_timeout_exception(self,
                                                         mock_request):
        params = {}
        mock_request.side_effect = requests.exceptions.ConnectTimeout()
        r, resp = self.client._send_request("POST",
                                            "/metrics/generate",
                                            params)
        self.assertEqual(500, r.status_code)

    @mock.patch("requests.request")
    def test_send_post_request_read_timeout_exception(self,
                                                      mock_request):
        params = {}
        mock_request.side_effect = requests.exceptions.ReadTimeout()
        r, resp = self.client._send_request("POST",
                                            "/metrics/generate",
                                            params)
        self.assertEqual(500, r.status_code)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from cinder import coordination
from cinder.volume import driver
from cinder.volume.drivers.san import san


class PowerMaxNVMEBaseDriver(san.SanDriver, driver.BaseVD):

    driver_prefix = 'powermax'

    def ensure_export(self, context, volume):
        pass

    def create_export(self, context, volume, connector):
        pass

    def remove_export(self, context, volume):
        pass

    @staticmethod
    def check_for_export(context, volume_id):
        """Make sure volume is exported.

        :param context: the context
        :param volume_id: the volume id
        """
        pass

    def check_for_setup_error(self):
        """Validate the Unisphere version."""
        pass

    def create_volume(self, volume):
        """Creates a PowerMax volume.

        :param volume: the cinder volume object
        :returns: provider location dict
        """
        return self.common.create_volume(volume)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        :param volume: the cinder volume object
        :param snapshot: the cinder snapshot object
        :returns: provider location dict
        """
        return self.common.create_volume_from_snapshot(
            volume, snapshot)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a cloned volume.

        :param volume: the cinder volume object
        :param src_vref: the source volume reference
        :returns: provider location dict
        """
        return self.common.create_cloned_volume(volume, src_vref)

    def delete_volume(self, volume):
        """Deletes a PowerMax volume.

        :param volume: the cinder volume object
        """
        self.common.delete_volume(volume)

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: the cinder snapshot object
        :returns: provider location dict
        """
        src_volume = snapshot.volume
        return self.common.create_snapshot(snapshot, src_volume)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: the cinder snapshot object
        """
        src_volume = snapshot.volume

        self.common.delete_snapshot(snapshot, src_volume)

    @coordination.synchronized('{self.driver_prefix}-{volume.id}')
    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info."""
        pass

    @coordination.synchronized('{self.driver_prefix}-{volume.id}')
    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector.

        :param volume: the volume object
        :param connector: the connector object
        """
        self.common.terminate_connection(volume, connector)

    def extend_volume(self, volume, new_size):
        """Extend an existing volume.

        :param volume: the cinder volume object
        :param new_size: the required new size
        """
        self.common.extend_volume(volume, new_size)

    def manage_existing(self, volume, existing_ref):
        """Manages an existing PowerMax Volume (import to Cinder).

        Renames the Volume to match the expected name for the volume.
        Also need to consider things like QoS, Emulation, account/tenant.
        """
        return self.common.manage_existing(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of an existing PowerMax volume to manage_existing.

        :param self: reference to class
        :param volume: the volume object including the volume_type_id
        :param external_ref: reference to the existing volume
        :returns: size of the volume in GB
        """
        return self.common.manage_existing_get_size(volume, existing_ref)

    def unmanage(self, volume):
        """Export PowerMax volume from Cinder.

        Leave the volume intact on the backend array.
        """
        return self.common.unmanage(volume)

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Manage an existing PowerMax Snapshot (import to Cinder).

        Renames the Snapshot to prefix it with OS- to indicate
        it is managed by Cinder.

        :param snapshot: the snapshot object
        :param existing_ref: the snapshot name on the backend PowerMax
        :returns: model_update
        """
        return self.common.manage_existing_snapshot(snapshot, existing_ref)

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return the size of the source volume for manage-existing-snapshot.

        :param snapshot: the snapshot object
        :param existing_ref: the snapshot name on the backend PowerMax
        :returns: size of the source volume in GB
        """
        return self.common.manage_existing_snapshot_get_size(snapshot)

    def unmanage_snapshot(self, snapshot):
        """Export PowerMax Snapshot from Cinder.

        Leaves the snapshot intact on the backend PowerMax.

        :param snapshot: the snapshot object
        """
        self.common.unmanage_snapshot(snapshot)

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """Lists all manageable volumes.

        :param cinder_volumes: List of currently managed Cinder volumes.
                               Unused in driver.
        :param marker: Begin returning volumes that appear later in the volume
                       list than that represented by this reference.
        :param limit: Maximum number of volumes to return. Default=1000.
        :param offset: Number of volumes to skip after marker.
        :param sort_keys: Results sort key. Valid keys: size, reference.
        :param sort_dirs: Results sort direction. Valid dirs: asc, desc.
        :returns: List of dicts containing all manageable volumes.
        """
        return self.common.get_manageable_volumes(marker, limit, offset,
                                                  sort_keys, sort_dirs)

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit, offset,
                                 sort_keys, sort_dirs):
        """Lists all manageable snapshots.

        :param cinder_snapshots: List of currently managed Cinder snapshots.
                                 Unused in driver.
        :param marker: Begin returning volumes that appear later in the
                       snapshot list than that represented by this reference.
        :param limit: Maximum number of snapshots to return. Default=1000.
        :param offset: Number of snapshots to skip after marker.
        :param sort_keys: Results sort key. Valid keys: size, reference.
        :param sort_dirs: Results sort direction. Valid dirs: asc, desc.
        :returns: List of dicts containing all manageable snapshots.
        """
        return self.common.get_manageable_snapshots(marker, limit, offset,
                                                    sort_keys, sort_dirs)

    def retype(self, context, volume, new_type, diff, host):
        """Migrate volume to another host using retype.

        :param context: context
        :param volume: the volume object including the volume_type_id
        :param new_type: the new volume type.
        :param diff: difference between old and new volume types.
            Unused in driver.
        :param host: the host dict holding the relevant
            target(destination) information
        :returns: boolean -- True if retype succeeded, False if error
        """
        return self.common.retype(volume, new_type, host)

    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Failover volumes to a secondary host/ backend.

        :param context: the context
        :param volumes: the list of volumes to be failed over
        :param secondary_id: the backend to be failed over to, is 'default'
                             if fail back
        :param groups: replication groups
        :returns: secondary_id, volume_update_list, group_update_list
        """
        active_backend_id, volume_update_list, group_update_list = (
            self.common.failover(volumes, secondary_id, groups))
        self.common.failover_completed(secondary_id, False)
        return active_backend_id, volume_update_list, group_update_list

    def failover(self, context, volumes, secondary_id=None, groups=None):
        """Like failover but for a host that is clustered."""
        return self.common.failover(volumes, secondary_id, groups)

    def failover_completed(self, context, active_backend_id=None):
        """This method is called after failover for clustered backends."""
        return self.common.failover_completed(active_backend_id, True)

    def create_group(self, context, group):
        """Creates a generic volume group.

        :param context: the context
        :param group: the group object
        :returns: model_update
        """
        return self.common.create_group(context, group)

    def delete_group(self, context, group, volumes):
        """Deletes a generic volume group.

        :param context: the context
        :param group: the group object
        :param volumes: the member volumes
        """
        return self.common.delete_group(
            context, group, volumes)

    def create_group_snapshot(self, context, group_snapshot, snapshots):
        """Creates a group snapshot.

        :param context: the context
        :param group_snapshot: the group snapshot
        :param snapshots: snapshots list
        """
        return self.common.create_group_snapshot(context,
                                                 group_snapshot, snapshots)

    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        """Deletes a group snapshot.

        :param context: the context
        :param group_snapshot: the grouop snapshot
        :param snapshots: snapshots list
        """
        return self.common.delete_group_snapshot(context,
                                                 group_snapshot, snapshots)

    def update_group(self, context, group,
                     add_volumes=None, remove_volumes=None):
        """Updates LUNs in group.

        :param context: the context
        :param group: the group object
        :param add_volumes: flag for adding volumes
        :param remove_volumes: flag for removing volumes
        """
        return self.common.update_group(group, add_volumes,
                                        remove_volumes)

    def create_group_from_src(
            self, context, group, volumes, group_snapshot=None,
            snapshots=None, source_group=None, source_vols=None):
        """Creates the volume group from source.

        :param context: the context
        :param group: the consistency group object to be created
        :param volumes: volumes in the group
        :param group_snapshot: the source volume group snapshot
        :param snapshots: snapshots of the source volumes
        :param source_group: the dictionary of a volume group as source.
        :param source_vols: a list of volume dictionaries in the source_group.
        """
        return self.common.create_group_from_src(
            context, group, volumes, group_snapshot, snapshots, source_group,
            source_vols)

    def enable_replication(self, context, group, volumes):
        """Enable replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :returns: model_update, None
        """
        return self.common.enable_replication(context, group, volumes)

    def disable_replication(self, context, group, volumes):
        """Disable replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :returns: model_update, None
        """
        return self.common.disable_replication(context, group, volumes)

    def failover_replication(self, context, group, volumes,
                             secondary_backend_id=None):
        """Failover replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :param secondary_backend_id: the secondary backend id - default None
        :returns: model_update, vol_model_updates
        """
        return self.common.failover_replication(
            context, group, volumes, secondary_backend_id)

    def revert_to_snapshot(self, context, volume, snapshot):
        """Revert volume to snapshot

        :param context: the context
        :param volume: the cinder volume object
        :param snapshot: the cinder snapshot object
        """
        self.common.revert_to_snapshot(volume, snapshot)

    @classmethod
    def clean_volume_file_locks(cls, volume_id):
        coordination.synchronized_remove(f'{cls.driver_prefix}-{volume_id}')



# Copyright (C) 2022, Hewlett Packard Enterprise, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Utility module for Hewlett Packard Enterprise Driver."""

VERSION = '1.0.0'
CI_WIKI_NAME = 'HPE_XP_Storage_CI'
PARAM_PREFIX = 'hpexp'
VENDOR_NAME = 'Hewlett Packard Enterprise'
DRIVER_PREFIX = 'HPEXP'
DRIVER_FILE_PREFIX = 'hpe_xp'
TARGET_PREFIX = 'HPEXP-'
HDP_VOL_ATTR = 'THP'
HDT_VOL_ATTR = 'ST'
NVOL_LDEV_TYPE = 'THP V-VOL'
TARGET_IQN_SUFFIX = '.hpexp-target'
PAIR_ATTR = 'FS'



# Copyright (c) 2012 NetApp, Inc.  All rights reserved.
# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Bob Callaway.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Contains configuration options for NetApp drivers.

Common place to hold configuration options for all NetApp drivers.
Options need to be grouped into granular units to be able to be reused
by different modules and classes. This does not restrict declaring options in
individual modules. If options are not re usable then can be declared in
individual modules. It is recommended to Keep options at a single
place to ensure re usability and better management of configuration options.
"""

from oslo_config import cfg
from oslo_config import types

from cinder.volume import configuration as conf

NETAPP_SIZE_MULTIPLIER_DEFAULT = 1.2

netapp_proxy_opts = [
    cfg.StrOpt('netapp_storage_family',
               default='ontap_cluster',
               choices=['ontap_cluster'],
               help=('The storage family type used on the storage system; '
                     'the only valid value is ontap_cluster for using '
                     'clustered Data ONTAP.')),
    cfg.StrOpt('netapp_storage_protocol',
               choices=['iscsi', 'fc', 'nfs', 'nvme'],
               help=('The storage protocol to be used on the data path with '
                     'the storage system.')), ]

netapp_connection_opts = [
    cfg.StrOpt('netapp_server_hostname',
               help='The hostname (or IP address) for the storage system or '
                    'proxy server.'),
    cfg.IntOpt('netapp_server_port',
               help=('The TCP port to use for communication with the storage '
                     'system or proxy server. If not specified, Data ONTAP '
                     'drivers will use 80 for HTTP and 443 for HTTPS.')),
    cfg.BoolOpt('netapp_use_legacy_client',
                default=True,
                help=('Select which ONTAP client to use for retrieving and '
                      'modifying data on the storage. The legacy client '
                      'relies on ZAPI calls. If set to False, the new REST '
                      'client is used, which runs REST calls if supported, '
                      'otherwise falls back to the equivalent ZAPI call.')),
    cfg.IntOpt('netapp_async_rest_timeout',
               min=60,
               default=60,  # One minute
               help='The maximum time in seconds to wait for completing a '
                    'REST asynchronous operation.'), ]

netapp_transport_opts = [
    cfg.StrOpt('netapp_transport_type',
               default='http',
               choices=['http', 'https'],
               help=('The transport protocol used when communicating with '
                     'the storage system or proxy server.')),
    cfg.StrOpt('netapp_ssl_cert_path',
               help=("The path to a CA_BUNDLE file or directory with "
                     "certificates of trusted CA. If set to a directory, it "
                     "must have been processed using the c_rehash utility "
                     "supplied with OpenSSL. If not informed, it will use the "
                     "Mozilla's carefully curated collection of Root "
                     "Certificates for validating the trustworthiness of SSL "
                     "certificates. Only applies with new REST client.")), ]

netapp_basicauth_opts = [
    cfg.StrOpt('netapp_login',
               help=('Administrative user account name used to access the '
                     'storage system or proxy server.')),
    cfg.StrOpt('netapp_password',
               help=('Password for the administrative user account '
                     'specified in the netapp_login option.'),
               secret=True), ]

netapp_certificateauth_opts = [
    cfg.StrOpt('netapp_private_key_file',
               sample_default='/path/to/private_key.key',
               help=("""
                     This option is applicable for both self signed and ca
                     verified certificates.

                     For self signed certificate: Absolute path to the file
                     containing the private key associated with the self
                     signed certificate. It is a sensitive file that should
                     be kept secure and protected. The private key is used
                     to sign the certificate and establish the authenticity
                     and integrity of the certificate during the
                     authentication process.

                     For ca verified certificate: Absolute path to the file
                     containing the private key associated with the
                     certificate. It is generated when creating the
                     certificate signingrequest (CSR) and should be kept
                     secure and protected. The private key is used to sign
                     the CSR and later used to establish secure connections
                     and authenticate the entity.
                     """),
               secret=True),
    cfg.StrOpt('netapp_certificate_file',
               sample_default='/path/to/certificate.pem',
               help=("""
                     This option is applicable for both self signed and ca
                     verified certificates.

                     For self signed certificate: Absolute path to the file
                     containing the self-signed digital certificate itself.
                     It includes information about the entity such as the
                     common name (e.g., domain name), organization details,
                     validity period, and public key. The certificate file
                     is generated based on the private key and is used by
                     clients or systems to verify the entity identity during
                     the authentication process.

                     For ca verified certificate: Absolute path to the file
                     containing the digital certificate issued by the
                     trusted third-party certificate authority (CA). It
                     includes information about the entity identity, public
                     key, and the CA that issued the certificate. The
                     certificate file is used by clients or systems to verify
                     the authenticity and integrity of the entity during the
                     authentication process.
                     """),
               secret=True),
    cfg.StrOpt('netapp_ca_certificate_file',
               sample_default='/path/to/ca_certificate.crt',
               help=("""
                     This option is applicable only for a ca verified
                     certificate.

                     Ca verified file: Absolute path to the file containing
                     the public key certificate of the trusted third-party
                     certificate authority (CA) that issued the certificate.
                     It is used by clients or systems to validate the
                     authenticity of the certificate presented by the
                     entity. The CA certificate file is typically pre
                     configured in the trust store of clients or systems to
                     establish trust in certificates issued by that CA.
                     """),
               secret=True),
    cfg.BoolOpt('netapp_certificate_host_validation',
                default=False,
                help=('This option is used only if netapp_private_key_file'
                      ' and netapp_certificate_file files are passed in the'
                      ' configuration.'
                      ' By default certificate verification is disabled'
                      ' and to verify the certificates please set the value'
                      ' to True.')), ]

netapp_provisioning_opts = [
    cfg.FloatOpt('netapp_size_multiplier',
                 default=NETAPP_SIZE_MULTIPLIER_DEFAULT,
                 help=('The quantity to be multiplied by the requested '
                       'volume size to ensure enough space is available on '
                       'the virtual storage server (Vserver) to fulfill '
                       'the volume creation request.  Note: this option '
                       'is deprecated and will be removed in favor of '
                       '"reserved_percentage" in the Mitaka release.')),
    cfg.StrOpt('netapp_lun_space_reservation',
               default='enabled',
               choices=['enabled', 'disabled'],
               help=('This option determines if storage space is reserved '
                     'for LUN allocation. If enabled, LUNs are thick '
                     'provisioned. If space reservation is disabled, '
                     'storage space is allocated on demand.')),
    cfg.BoolOpt('netapp_driver_reports_provisioned_capacity',
                default=False,
                help=('Set to True for Cinder to query the storage system in '
                      'order to calculate volumes provisioned size, otherwise '
                      'provisioned_capacity_gb will corresponds to the '
                      'value of allocated_capacity_gb (calculated by Cinder '
                      'Core code). Enabling this feature increases '
                      'the number of API calls to the storage and '
                      'requires more processing on host, which may impact '
                      'volume report overall performance.')), ]

netapp_cluster_opts = [
    cfg.StrOpt('netapp_vserver',
               help=('This option specifies the virtual storage server '
                     '(Vserver) name on the storage cluster on which '
                     'provisioning of block storage volumes should occur.')),
    cfg.BoolOpt('netapp_disaggregated_platform',
                default=False,
                help=('This option specifies whether to enable ASA r2 '
                      'workflows for NetApp disaggregated platform. '
                      'When set to True, the driver will use ASA r2 '
                      'specific client and workflows for interacting '
                      'with NetApp ONTAP.')), ]

netapp_img_cache_opts = [
    cfg.IntOpt('netapp_nfs_image_cache_cleanup_interval',
               default=600,
               min=60,
               help=('Sets time in seconds between NFS image cache '
                     'cleanup tasks.')),
    cfg.IntOpt('thres_avl_size_perc_start',
               default=20,
               help=('If the percentage of available space for an NFS share '
                     'has dropped below the value specified by this option, '
                     'the NFS image cache will be cleaned.')),
    cfg.IntOpt('thres_avl_size_perc_stop',
               default=60,
               help=('When the percentage of available space on an NFS share '
                     'has reached the percentage specified by this option, '
                     'the driver will stop clearing files from the NFS image '
                     'cache that have not been accessed in the last M '
                     'minutes, where M is the value of the '
                     'expiry_thres_minutes configuration option.')),
    cfg.IntOpt('expiry_thres_minutes',
               default=720,
               help=('This option specifies the threshold for last access '
                     'time for images in the NFS image cache. When a cache '
                     'cleaning cycle begins, images in the cache that have '
                     'not been accessed in the last M minutes, where M is '
                     'the value of this parameter, will be deleted from the '
                     'cache to create free space on the NFS share.')), ]

netapp_nfs_extra_opts = [
    cfg.StrOpt('netapp_copyoffload_tool_path',
               help=('This option specifies the path of the NetApp copy '
                     'offload tool binary. Ensure that the binary has execute '
                     'permissions set which allow the effective user of the '
                     'cinder-volume process to execute the file.'),
               deprecated_for_removal=True,
               deprecated_reason='The CopyOfflload tool is no longer '
                                 'available for downloading.'), ]
netapp_san_opts = [
    cfg.StrOpt('netapp_lun_ostype',
               help=('This option defines the type of operating system that'
                     ' will access a LUN exported from Data ONTAP; it is'
                     ' assigned to the LUN at the time it is created.')),
    cfg.StrOpt('netapp_namespace_ostype',
               help=('This option defines the type of operating system that'
                     ' will access a namespace exported from Data ONTAP; it is'
                     ' assigned to the namespace at the time it is created.')),
    cfg.StrOpt('netapp_host_type',
               help=('This option defines the type of operating system for'
                     ' all initiators that can access a LUN. This information'
                     ' is used when mapping LUNs to individual hosts or'
                     ' groups of hosts.')),
    cfg.StrOpt('netapp_pool_name_search_pattern',
               deprecated_opts=[cfg.DeprecatedOpt(name='netapp_volume_list'),
                                cfg.DeprecatedOpt(name='netapp_storage_pools')
                                ],
               default="(.+)",
               help=('This option is used to restrict provisioning to the '
                     'specified pools. Specify the value of '
                     'this option to be a regular expression which will be '
                     'applied to the names of objects from the storage '
                     'backend which represent pools in Cinder. This option '
                     'is only utilized when the storage protocol is '
                     'configured to use iSCSI or FC.')),
    cfg.IntOpt('netapp_lun_clone_busy_timeout',
               min=0,
               default=30,
               help='Specifies the maximum time (in seconds) to retry'
                    ' the LUN clone operation when an ONTAP "device busy"'
                    ' error occurs.'),
    cfg.IntOpt('netapp_lun_clone_busy_interval',
               min=0,
               default=3,
               help='Specifies the time interval (in seconds) to retry'
                    ' the LUN clone operation when an ONTAP "device busy"'
                    ' error occurs.')]

netapp_replication_opts = [
    cfg.MultiOpt('netapp_replication_aggregate_map',
                 item_type=types.Dict(),
                 help="Multi opt of dictionaries to represent the aggregate "
                      "mapping between source and destination back ends when "
                      "using whole back end replication. For every "
                      "source aggregate associated with a cinder pool (NetApp "
                      "FlexVol/FlexGroup), you would need to specify the "
                      "destination aggregate on the replication target "
                      "device. "
                      "A replication target device is configured with the "
                      "configuration option replication_device. Specify this "
                      "option as many times as you have replication devices. "
                      "Each entry takes the standard dict config form: "
                      "netapp_replication_aggregate_map = "
                      "backend_id:<name_of_replication_device_section>,"
                      "src_aggr_name1:dest_aggr_name1,"
                      "src_aggr_name2:dest_aggr_name2,..."),
    cfg.IntOpt('netapp_snapmirror_quiesce_timeout',
               min=0,
               default=3600,  # One Hour
               help='The maximum time in seconds to wait for existing '
                    'SnapMirror transfers to complete before aborting '
                    'during a failover.'),
    cfg.IntOpt('netapp_replication_volume_online_timeout',
               min=60,
               default=360,  # Default to six minutes
               help='Sets time in seconds to wait for a replication volume '
                    'create to complete and go online.'),
    cfg.StrOpt('netapp_replication_policy',
               default='MirrorAllSnapshots',
               help='This option defines the replication policy to be used '
                    'while creating snapmirror relationship. Default is '
                    'MirrorAllSnapshots which is based on async-mirror.'
                    'User can pass values like Sync, StrictSync for '
                    'synchronous snapmirror relationship (SM-S) to achieve '
                    'zero RPO')]

netapp_support_opts = [
    cfg.StrOpt('netapp_api_trace_pattern',
               default='(.*)',
               help=('A regular expression to limit the API tracing. This '
                     'option is honored only if enabling ``api`` tracing '
                     'with the ``trace_flags`` option. By default, '
                     'all APIs will be traced.')),
]

netapp_migration_opts = [
    cfg.IntOpt('netapp_migrate_volume_timeout',
               default=3600,
               min=30,
               help='Sets time in seconds to wait for storage assisted volume '
                    'migration to complete.'),
]

CONF = cfg.CONF
CONF.register_opts(netapp_proxy_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_connection_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_transport_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_basicauth_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_certificateauth_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_cluster_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_provisioning_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_img_cache_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_nfs_extra_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_san_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_replication_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_support_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_migration_opts, group=conf.SHARED_CONF_GROUP)



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""iSCSI volume driver for JovianDSS driver."""
import math
import string

from oslo_log import log as logging
from oslo_utils import units as o_units

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.open_e.jovian_common import driver as jdriver
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom
from cinder.volume.drivers.open_e.jovian_common import rest
from cinder.volume.drivers.open_e import options
from cinder.volume.drivers.san import san
from cinder.volume import volume_utils


LOG = logging.getLogger(__name__)


@interface.volumedriver
class JovianISCSIDriver(driver.ISCSIDriver):
    """Executes volume driver commands on Open-E JovianDSS.

    Version history:

    .. code-block:: none

        1.0.0 - Open-E JovianDSS driver with basic functionality
        1.0.1 - Added certificate support
                Added revert to snapshot support
        1.0.2 - Added multi-attach support
                Added 16K block support
        1.0.3 - Driver rework and optimisation
                Abandon recursive volume deletion
                Removed revert to snapshot support
 """

    # Third-party Systems wiki page
    CI_WIKI_NAME = "Open-E_JovianDSS_CI"
    VERSION = "1.0.3"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._stats = None
        self.jovian_iscsi_target_portal_port = "3260"
        self.jovian_target_prefix = 'iqn.2020-04.com.open-e.cinder:'
        self.jovian_chap_pass_len = 12
        self.jovian_sparse = False
        self.jovian_ignore_tpath = None
        self.jovian_hosts = None
        self._pool = 'Pool-0'
        self.ra = None
        self.driver = None

    @property
    def backend_name(self):
        """Return backend name."""
        backend_name = None
        if self.configuration:
            backend_name = self.configuration.get('volume_backend_name',
                                                  'Open-EJovianDSS')
        if not backend_name:
            backend_name = self.__class__.__name__
        return backend_name

    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        self.configuration.append_config_values(
            options.jdss_connection_opts)
        self.configuration.append_config_values(
            options.jdss_iscsi_opts)
        self.configuration.append_config_values(
            options.jdss_volume_opts)
        self.configuration.append_config_values(san.san_opts)

        self.jovian_target_prefix = self.configuration.get(
            'target_prefix',
            'iqn.2020-04.com.open-e.cinder:')
        self.jovian_chap_pass_len = self.configuration.get(
            'chap_password_len', 12)
        self.block_size = (
            self.configuration.get('jovian_block_size', '64K'))
        self.jovian_sparse = (
            self.configuration.get('san_thin_provision', True))
        self.jovian_ignore_tpath = self.configuration.get(
            'jovian_ignore_tpath', None)
        self.jovian_hosts = self.configuration.get(
            'san_hosts', [])

        self.ra = rest.JovianRESTAPI(self.configuration)
        self.driver = jdriver.JovianDSSDriver(self.configuration)

    def check_for_setup_error(self):
        """Check for setup error."""
        if len(self.jovian_hosts) == 0:
            msg = _("No hosts provided in configuration")
            raise exception.VolumeDriverException(msg)

        if not self.driver.rest_config_is_ok():
            msg = (_("Unable to identify pool %s") % self._pool)
            raise exception.VolumeDriverException(msg)

        valid_bsize = ['16K', '32K', '64K', '128K', '256K', '512K', '1M']
        if self.block_size not in valid_bsize:
            raise exception.InvalidConfigurationValue(
                value=self.block_size,
                option='jovian_block_size')

    def _get_target_name(self, volume_name):
        """Return iSCSI target name to access volume."""
        return f'{self.jovian_target_prefix}{volume_name}'

    def _get_active_ifaces(self):
        """Return list of ip addresses for iSCSI connection"""

        return self.jovian_hosts

    def create_volume(self, volume):
        """Create a volume.

        :param volume: volume reference
        :return: model update dict for volume reference
        """
        LOG.debug('creating volume %s.', volume.id)

        try:
            self.driver.create_volume(volume.id,
                                      volume.size,
                                      sparse=self.jovian_sparse,
                                      block_size=self.block_size)

        except jexc.JDSSException as jerr:
            LOG.error("Create volume error. Because %(err)s",
                      {"err": jerr})
            raise exception.VolumeBackendAPIException(
                _('Failed to create volume %s.') % volume.id) from jerr

        return self._get_provider_info(volume.id)

    def delete_volume(self, volume, cascade=False):
        """Delete volume

        :param volume: volume reference
        :param cascade: remove snapshots of a volume as well
        """

        try:
            self.driver.delete_volume(volume.id, cascade=cascade)
        except jexc.JDSSException as jerr:
            raise exception.VolumeBackendAPIException(jerr)

    def extend_volume(self, volume, new_size):
        """Extend an existing volume.

        :param volume: volume reference
        :param new_size: volume new size in GB
        """

        try:
            self.driver.resize_volume(volume.id, new_size)
        except jexc.JDSSException as jerr:
            msg = _('Failed to extend volume %s.')
            raise exception.VolumeBackendAPIException(
                data=msg % volume.id) from jerr

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the specified volume.

        :param volume: new volume reference
        :param src_vref: source volume reference
        """

        try:
            self.driver.create_cloned_volume(volume.id,
                                             src_vref.id,
                                             volume.size,
                                             sparse=self.jovian_sparse)
        except jexc.JDSSException as jerr:
            msg = _("Fail to clone volume %(vol)s to %(clone)s because of "
                    "error %(err)s.") % {
                'vol': src_vref.id,
                'clone': volume.id,
                'err': jerr}
            raise exception.VolumeBackendAPIException(msg) from jerr

        return self._get_provider_info(volume.id)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot.

        """
        LOG.debug('create volume %(vol)s from snapshot %(snap)s', {
            'vol': volume.id,
            'snap': snapshot.id})

        try:
            self.driver.create_cloned_volume(volume.id,
                                             snapshot.volume_id,
                                             volume.size,
                                             snapshot_name=snapshot.id)
        except jexc.JDSSResourceExistsException as jerr:
            raise exception.Duplicate() from jerr
        except jexc.JDSSException as jerr:
            raise exception.VolumeBackendAPIException(
                _("Failed to create clone %(clone)s from snapshot %(snap)s "
                  "of volume %(vol)s because of error %(err)s.") % {
                    'vol': snapshot.volume_id,
                    'clone': volume.id,
                    'snap': snapshot.id,
                    'err': jerr}) from jerr

        return self._get_provider_info(volume.id)

    def create_snapshot(self, snapshot):
        """Create snapshot of existing volume.

        :param snapshot: snapshot object
        """
        LOG.debug('create snapshot %(snap)s for volume %(vol)s', {
            'snap': snapshot.id,
            'vol': snapshot.volume_id})

        try:
            self.driver.create_snapshot(snapshot.id, snapshot.volume_id)
        except jexc.JDSSSnapshotExistsException as jexistserr:
            raise exception.Duplicate() from jexistserr
        except jexc.JDSSResourceNotFoundException as jerrnotfound:
            raise exception.VolumeNotFound(
                volume_id=jcom.idname(snapshot.volume_id)) from jerrnotfound
        except jexc.JDSSException as jerr:
            args = {'snapshot': snapshot.id,
                    'object': snapshot.volume_id,
                    'err': jerr}
            msg = (_('Failed to create tmp snapshot %(snapshot)s '
                     'for object %(object)s: %(err)s') % args)
            raise exception.VolumeBackendAPIException(msg) from jerr

        return self._get_provider_info(snapshot.id)

    def delete_snapshot(self, snapshot):
        """Delete snapshot of existing volume.

        :param snapshot: snapshot reference
        """

        try:
            self.driver.delete_snapshot(snapshot.volume_id, snapshot.id)
        except jexc.JDSSResourceNotFoundException:
            return
        except jexc.JDSSException as jerr:
            raise exception.VolumeBackendAPIException(jerr)

    def _get_provider_info(self, vid):
        '''returns provider info dict

        :param vid: volume id
        '''

        info = {}
        try:
            info['provider_location'] = self.driver.get_provider_location(vid)
        except jexc.JDSSException as jerr:
            msg = _("Fail to identify critical properties of "
                    "new volume %s.") % vid
            raise exception.VolumeBackendAPIException(data=msg) from jerr

        info['provider_auth'] = self._get_provider_auth()

        return info

    def _get_provider_auth(self):
        """Get provider authentication for the volume.

        :return: string of auth method and credentials
        """
        chap_user = volume_utils.generate_password(
            length=8,
            symbolgroups=(string.ascii_lowercase +
                          string.ascii_uppercase))

        chap_password = volume_utils.generate_password(
            length=self.jovian_chap_pass_len,
            symbolgroups=(string.ascii_lowercase +
                          string.ascii_uppercase + string.digits))

        return 'CHAP %(user)s %(passwd)s' % {
            'user': chap_user, 'passwd': chap_password}

    def create_export(self, _ctx, volume, connector):
        """Create new export for zvol.

        :param volume: reference of volume to be exported
        :return: iscsiadm-formatted provider location string
        """
        LOG.debug("create export for volume: %s.", volume.id)

        provider_auth = volume.provider_auth
        ret = dict()

        if provider_auth is None:
            provider_auth = self._get_provider_auth()
            ret['provider_auth'] = provider_auth

        try:
            self.driver.ensure_export(volume.id, provider_auth)
            location = self.driver.get_provider_location(volume.id)
            ret['provider_location'] = location
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr

        return ret

    def ensure_export(self, _ctx, volume):
        """Recreate parts of export if necessary.

        :param volume: reference of volume to be exported
        """
        LOG.debug("ensure export for volume: %s.", volume.id)
        provider_auth = volume.provider_auth
        ret = dict()

        if provider_auth is None:
            provider_auth = self._get_provider_auth()
            ret['provider_auth'] = provider_auth
        try:
            self.driver.ensure_export(volume.id, provider_auth)
            location = self.driver.get_provider_location(volume.id)
            ret['provider_location'] = location
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr
        return ret

    def create_export_snapshot(self, context, snapshot, connector):
        provider_auth = snapshot.provider_auth
        ret = dict()

        if provider_auth is None:
            provider_auth = self._get_provider_auth()
            ret['provider_auth'] = provider_auth
        try:
            ret = self.driver.create_export_snapshot(snapshot.id,
                                                     snapshot.volume_id,
                                                     provider_auth)
        except jexc.JDSSResourceExistsException as jres_err:
            raise exception.Duplicate() from jres_err
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr
        return ret

    def remove_export(self, _ctx, volume):
        """Destroy all resources created to export zvol.

        :param volume: reference of volume to be unexposed
        """
        LOG.debug("remove_export for volume: %s.", volume.id)

        try:
            self.driver.remove_export(volume.id)
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr

    def remove_export_snapshot(self, context, snapshot):
        try:
            self.driver.remove_export_snapshot(snapshot.id,
                                               snapshot.volume_id)
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr

    def _update_volume_stats(self):
        """Retrieve stats info."""
        LOG.debug('Updating volume stats')

        pool_stats = self.ra.get_pool_stats()
        total_capacity = math.floor(int(pool_stats["size"]) / o_units.Gi)
        free_capacity = math.floor(int(pool_stats["available"]) / o_units.Gi)

        reserved_percentage = (
            self.configuration.get('reserved_percentage', 0))

        if total_capacity is None:
            total_capacity = 'unknown'
        if free_capacity is None:
            free_capacity = 'unknown'

        location_info = '%(driver)s:%(host)s:%(volume)s' % {
            'driver': self.__class__.__name__,
            'host': self.ra.get_active_host(),
            'volume': self._pool
        }

        self._stats = {
            'vendor_name': 'Open-E',
            'driver_version': self.VERSION,
            'storage_protocol': constants.ISCSI,
            'total_capacity_gb': total_capacity,
            'free_capacity_gb': free_capacity,
            'reserved_percentage': int(reserved_percentage),
            'volume_backend_name': self.backend_name,
            'QoS_support': False,
            'location_info': location_info,
            'multiattach': True
        }

        LOG.debug('Total capacity: %d, '
                  'Free %d.',
                  self._stats['total_capacity_gb'],
                  self._stats['free_capacity_gb'])

    def _get_iscsi_properties(self, volume_id, provider_auth,
                              multipath=False):
        """Return dict according to cinder/driver.py implementation.

        :param volume_id: openstack volume UUID
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        :param bool multipath: use multipath flag
        :return:
        """
        tname = self.jovian_target_prefix + volume_id
        iface_info = []
        if multipath:
            iface_info = self._get_active_ifaces()
            if not iface_info:
                raise exception.InvalidConfigurationValue(
                    _('No available interfaces '
                      'or config excludes them'))

        iscsi_properties = {}

        if multipath:
            iscsi_properties['target_iqns'] = []
            iscsi_properties['target_portals'] = []
            iscsi_properties['target_luns'] = []
            LOG.debug('tpaths %s.', iface_info)
            for iface in iface_info:
                iscsi_properties['target_iqns'].append(
                    self.jovian_target_prefix +
                    volume_id)
                iscsi_properties['target_portals'].append(
                    iface +
                    ":" +
                    str(self.jovian_iscsi_target_portal_port))
                iscsi_properties['target_luns'].append(0)
        else:
            iscsi_properties['target_iqn'] = tname
            iscsi_properties['target_portal'] = (
                self.ra.get_active_host() +
                ":" +
                str(self.jovian_iscsi_target_portal_port))

        iscsi_properties['target_discovered'] = False

        if provider_auth is None:
            provider_auth = self._get_provider_auth()

        (auth_method, auth_username, auth_secret) = provider_auth.split()

        iscsi_properties['auth_method'] = auth_method
        iscsi_properties['auth_username'] = auth_username
        iscsi_properties['auth_password'] = auth_secret

        iscsi_properties['target_lun'] = 0
        return iscsi_properties

    def initialize_connection(self, volume, connector):
        """Initialize the connection and returns connection info.

        The iscsi driver returns a driver_volume_type of 'iscsi'.
        Format of the driver data is defined in _get_iscsi_properties.
        Example return value:
        .. code-block:: json
            {
                'driver_volume_type': 'iscsi'
                'data': {
                    'target_discovered': True,
                    'target_iqn': 'iqn.2010-10.org.openstack:volume-00000001',
                    'target_portal': '127.0.0.0.1:3260',
                    'volume_id': '12345678-1234-1234-1234-123456789012',
                }
            }
        """
        multipath = connector.get("multipath", False)

        provider_auth = volume.provider_auth

        ret = {
            'driver_volume_type': 'iscsi',
            'data': None,
        }

        try:
            self.driver.initialize_connection(volume.id,
                                              provider_auth,
                                              multipath=multipath)
            ret['data'] = self._get_iscsi_properties(volume.id,
                                                     provider_auth,
                                                     multipath=multipath)
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr
        return ret

    def terminate_connection(self, volume, connector, force=False, **kwargs):
        """terminate_connection

        """

        LOG.debug("terminate connection for %(volume)s ",
                  {'volume': volume.id})

    def initialize_connection_snapshot(self, snapshot, connector, **kwargs):
        multipath = connector.get("multipath", False)

        provider_auth = snapshot.provider_auth

        ret = {
            'driver_volume_type': 'iscsi',
            'data': None,
        }

        try:
            self.driver.initialize_connection(snapshot.volume_id,
                                              provider_auth,
                                              snapshot_id=snapshot.id,
                                              multipath=multipath)
            ret['data'] = self._get_iscsi_properties(snapshot.id,
                                                     provider_auth,
                                                     multipath=multipath)
        except jexc.JDSSException as jerr:
            raise exception.VolumeDriverException from jerr

        return ret

    def terminate_connection_snapshot(self, snapshot, connector, **kwargs):
        pass



#  Copyright, 2023, KNS Group LLC (YADRO)
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import log as logging

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.yadro import tatlin_common
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class TatlinFCVolumeDriver(tatlin_common.TatlinCommonVolumeDriver,
                           driver.FibreChannelDriver):
    """ACCESS Tatlin FC Driver.

    Executes commands relating to FC.
    Supports creation of volumes.

    .. code-block:: none

     API version history:

        1.0 - Initial version.
    """

    VERSION = '1.0'

    SUPPORTS_ACTIVE_ACTIVE = True

    # ThirdPartySystems wiki
    CI_WIKI_NAME = "Yadro_Tatlin_Unified_CI"

    def __init__(self, *args, **kwargs):
        super(TatlinFCVolumeDriver, self).__init__(*args, **kwargs)
        self.backend_name = self.configuration.safe_get(
            'volume_backend_name') or 'TatlinFC'
        self.DRIVER_VOLUME_TYPE = constants.FC
        self._lookup_service = fczm_utils.create_lookup_service()

    def _create_connection_info(self, volume, connector):
        info = {
            'driver_volume_type': constants.FC_VARIANT_1,
            'data': self._create_volume_data(volume, connector)
        }
        return info

    def _get_ports_portals(self):
        result = self.tatlin_api.get_port_portal("fc")
        ports = {}
        for p in result:
            iface = p['params']['ifname']
            if self._export_ports and iface not in self._export_ports:
                continue
            ports.setdefault(iface, []).append(p['params']['wwpn'])
        return ports

    def _create_volume_data(self, volume, connector):
        if connector is None:
            return {}
        lun_id = self._find_mapped_lun(volume.name_id, connector)
        volume_ports = self.tatlin_api.get_volume_ports(volume.name_id)
        ports_portals = self._get_ports_portals()
        data = {
            'target_discovered': True,
            'target_lun': lun_id,
            'discard': False,
        }
        target_wwns = []
        for port in volume_ports:
            wwpns = ports_portals.get(port['port'])
            if not wwpns:
                continue
            target_wwns += [w.replace(':', '') for w in wwpns]

        data['target_wwn'] = target_wwns
        data['initiator_target_map'] = self._build_initiator_target_map(
            target_wwns, connector)
        return data

    def find_current_host(self, connector):
        wwns = connector['wwpns']
        LOG.debug('Try to find host id for %s', wwns)
        result = self.tatlin_api.get_all_hosts()
        for h in result:
            for wwn in h['initiators']:
                if wwn.replace(':', '') in wwns:
                    LOG.debug('Current host is %s', h['id'])
                    return h['id']
        message = _('Unable to get host information for wwns: %s') % str(wwns)
        LOG.error(message)
        raise exception.VolumeBackendAPIException(message=message)

    def _build_initiator_target_map(self, target_wwns, connector):
        result = {}

        if self._lookup_service:
            mapping = self._lookup_service.get_device_mapping_from_network(
                connector['wwpns'], target_wwns)
            for fabric in mapping.values():
                for initiator in fabric['initiator_port_wwn_list']:
                    result.setdefault(initiator, []).extend(
                        fabric['target_port_wwn_list'])
            result = {i: list(set(t)) for i, t in result.items()}
        else:
            result = dict.fromkeys(connector['wwpns'], target_wwns)

        return result



# Copyright (C) 2022, Hewlett Packard Enterprise, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Fibre channel module for Hewlett Packard Enterprise Driver."""

from cinder import interface
from cinder.volume.drivers.hitachi import hbsd_common
from cinder.volume.drivers.hitachi import hbsd_fc
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_utils
from cinder.volume.drivers.hpe.xp import hpe_xp_rest as rest
from cinder.volume.drivers.hpe.xp import hpe_xp_utils as utils

MSG = hbsd_utils.HBSDMsg

_DRIVER_INFO = {
    'version': utils.VERSION,
    'proto': 'FC',
    'hba_id': 'wwpns',
    'hba_id_type': 'World Wide Name',
    'msg_id': {
        'target': MSG.CREATE_HOST_GROUP_FAILED,
    },
    'volume_backend_name': '%(prefix)sFC' % {
        'prefix': utils.DRIVER_PREFIX,
    },
    'volume_type': 'fibre_channel',
    'param_prefix': utils.PARAM_PREFIX,
    'vendor_name': utils.VENDOR_NAME,
    'driver_prefix': utils.DRIVER_PREFIX,
    'driver_file_prefix': utils.DRIVER_FILE_PREFIX,
    'target_prefix': utils.TARGET_PREFIX,
    'hdp_vol_attr': utils.HDP_VOL_ATTR,
    'hdt_vol_attr': utils.HDT_VOL_ATTR,
    'nvol_ldev_type': utils.NVOL_LDEV_TYPE,
    'target_iqn_suffix': utils.TARGET_IQN_SUFFIX,
    'pair_attr': utils.PAIR_ATTR,
}


@interface.volumedriver
class HPEXPFCDriver(hbsd_fc.HBSDFCDriver):
    """Fibre channel class for  Hewlett Packard Enterprise Driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver.

    """

    VERSION = utils.VERSION

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = utils.CI_WIKI_NAME

    def __init__(self, *args, **kwargs):
        """Initialize instance variables."""
        super(HPEXPFCDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(rest.COMMON_VOLUME_OPTS)
        self.configuration.append_config_values(rest.FC_VOLUME_OPTS)

    def _init_common(self, conf, db):
        return rest.HPEXPRESTFC(conf, _DRIVER_INFO, db)

    @staticmethod
    def get_driver_options():
        additional_opts = HPEXPFCDriver._get_oslo_driver_opts(
            *(hbsd_common._INHERITED_VOLUME_OPTS +
              hbsd_rest._REQUIRED_REST_OPTS +
              ['driver_ssl_cert_verify', 'driver_ssl_cert_path',
               'san_api_port', ]))
        return (rest.COMMON_VOLUME_OPTS +
                rest.REST_VOLUME_OPTS +
                rest.FC_VOLUME_OPTS +
                additional_opts)



# Copyright 2015 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import inspect
from unittest import mock

import tooz.coordination
import tooz.locking

from cinder import coordination
from cinder.tests.unit import test


class Locked(Exception):
    pass


class MockToozLock(tooz.locking.Lock):
    active_locks = set()

    def acquire(self, blocking=True):
        if self.name not in self.active_locks:
            self.active_locks.add(self.name)
            return True
        elif not blocking:
            return False
        else:
            raise Locked

    def release(self):
        self.active_locks.remove(self.name)


class CoordinatorTestCase(test.TestCase):
    MOCK_TOOZ = False

    @mock.patch('cinder.coordination.cfg.CONF.coordination.backend_url')
    @mock.patch('cinder.coordination.Coordinator._get_file_path')
    @mock.patch('tooz.coordination.get_coordinator')
    def test_coordinator_start(self, get_coordinator, mock_get_file_path,
                               mock_backend_url):
        crd = get_coordinator.return_value

        agent = coordination.Coordinator()
        self.assertIsNone(agent._file_path)
        agent.start()
        self.assertTrue(get_coordinator.called)
        self.assertTrue(crd.start.called)

        agent.start()
        crd.start.assert_called_once_with(start_heart=True)

        mock_get_file_path.assert_called_once_with(mock_backend_url)
        self.assertEqual(mock_get_file_path.return_value, agent._file_path)

    @mock.patch('tooz.coordination.get_coordinator')
    def test_coordinator_stop(self, get_coordinator):
        crd = get_coordinator.return_value

        agent = coordination.Coordinator()
        agent.start()
        self.assertIsNotNone(agent.coordinator)
        agent.stop()
        self.assertTrue(crd.stop.called)
        self.assertIsNone(agent.coordinator)

        agent.stop()
        crd.stop.assert_called_once_with()

    @mock.patch('tooz.coordination.get_coordinator')
    def test_coordinator_lock(self, get_coordinator):
        crd = get_coordinator.return_value
        crd.get_lock.side_effect = lambda n: MockToozLock(n)

        agent1 = coordination.Coordinator()
        agent1.start()
        agent2 = coordination.Coordinator()
        agent2.start()

        lock_name = 'lock'
        expected_name = lock_name.encode('ascii')

        self.assertNotIn(expected_name, MockToozLock.active_locks)
        with agent1.get_lock(lock_name):
            self.assertIn(expected_name, MockToozLock.active_locks)
            self.assertRaises(Locked, agent1.get_lock(lock_name).acquire)
            self.assertRaises(Locked, agent2.get_lock(lock_name).acquire)
        self.assertNotIn(expected_name, MockToozLock.active_locks)

    @mock.patch('tooz.coordination.get_coordinator')
    def test_coordinator_offline(self, get_coordinator):
        crd = get_coordinator.return_value
        crd.start.side_effect = tooz.coordination.ToozConnectionError('err')

        agent = coordination.Coordinator()
        self.assertRaises(tooz.coordination.ToozError, agent.start)
        self.assertFalse(agent.started)

    def test_get_file_path(self):
        backend_url = 'file:///opt/stack/data/cinder'
        res = coordination.COORDINATOR._get_file_path(backend_url)
        self.assertEqual('/opt/stack/data/cinder/cinder-', res)

    def test_get_file_path_non_file(self):
        backend_url = 'etcd3+http://192.168.1.95:2379'
        res = coordination.COORDINATOR._get_file_path(backend_url)
        self.assertIsNone(res)

    @mock.patch('cinder.coordination.COORDINATOR._file_path', None)
    @mock.patch('glob.glob')
    @mock.patch('os.remove')
    def test_remove_lock_non_file_lock(self, mock_remove, mock_glob):
        coordination.COORDINATOR.remove_lock('lock-file')
        mock_glob.assert_not_called()
        mock_remove.assert_not_called()

    @mock.patch('cinder.coordination.COORDINATOR._file_path', '/data/cinder-')
    @mock.patch('glob.glob')
    @mock.patch('os.remove')
    def test_remove_lock(self, mock_remove, mock_glob):
        mock_glob.return_value = ['/data/cinder-attachment_update-UUID-1',
                                  '/data/cinder-attachment_update-UUID-2']

        coordination.COORDINATOR.remove_lock('attachment_update-UUID-*')

        mock_glob.assert_called_once_with(
            '/data/cinder-attachment_update-UUID-*')
        self.assertEqual(2, mock_remove.call_count)
        mock_remove.assert_has_calls(
            [mock.call('/data/cinder-attachment_update-UUID-1'),
             mock.call('/data/cinder-attachment_update-UUID-2')])

    @mock.patch('cinder.coordination.COORDINATOR._file_path', '/data/cinder-')
    @mock.patch('cinder.coordination.LOG.warning')
    @mock.patch('glob.glob')
    @mock.patch('os.remove')
    def test_remove_lock_missing_file(self, mock_remove, mock_glob, mock_log):
        mock_glob.return_value = ['/data/cinder-attachment_update-UUID-1',
                                  '/data/cinder-attachment_update-UUID-2']
        mock_remove.side_effect = [OSError(errno.ENOENT, ''), None]

        coordination.COORDINATOR.remove_lock('attachment_update-UUID-*')

        mock_glob.assert_called_once_with(
            '/data/cinder-attachment_update-UUID-*')
        self.assertEqual(2, mock_remove.call_count)
        mock_remove.assert_has_calls(
            [mock.call('/data/cinder-attachment_update-UUID-1'),
             mock.call('/data/cinder-attachment_update-UUID-2')])
        mock_log.assert_not_called()

    @mock.patch('cinder.coordination.COORDINATOR._file_path', '/data/cinder-')
    @mock.patch('cinder.coordination.LOG.warning')
    @mock.patch('glob.glob')
    @mock.patch('os.remove')
    def test_remove_lock_unknown_failure(self, mock_remove, mock_glob,
                                         mock_log):
        mock_glob.return_value = ['/data/cinder-attachment_update-UUID-1',
                                  '/data/cinder-attachment_update-UUID-2']
        mock_remove.side_effect = [ValueError(), None]

        coordination.COORDINATOR.remove_lock('attachment_update-UUID-*')

        mock_glob.assert_called_once_with(
            '/data/cinder-attachment_update-UUID-*')
        self.assertEqual(2, mock_remove.call_count)
        mock_remove.assert_has_calls(
            [mock.call('/data/cinder-attachment_update-UUID-1'),
             mock.call('/data/cinder-attachment_update-UUID-2')])
        self.assertEqual(1, mock_log.call_count)


class CoordinationTestCase(test.TestCase):
    @mock.patch.object(coordination.COORDINATOR, 'get_lock')
    def test_synchronized(self, get_lock):
        @coordination.synchronized('lock-{f_name}-{foo.val}-{bar[val]}')
        def func(foo, bar):
            pass

        foo = mock.Mock()
        foo.val = 7
        bar = mock.MagicMock()
        bar.__getitem__.return_value = 8
        func(foo, bar)
        get_lock.assert_called_with('lock-func-7-8')
        self.assertEqual(['foo', 'bar'], inspect.getfullargspec(func)[0])

    @mock.patch('cinder.coordination.COORDINATOR.remove_lock')
    def test_synchronized_remove(self, mock_remove):
        coordination.synchronized_remove(mock.sentinel.glob_name)
        mock_remove.assert_called_once_with(mock.sentinel.glob_name)

    @mock.patch('cinder.coordination.COORDINATOR.remove_lock')
    def test_synchronized_remove_custom_coordinator(self, mock_remove):
        coordinator = mock.Mock()
        coordination.synchronized_remove(mock.sentinel.glob_name, coordinator)
        coordinator.remove_lock.assert_called_once_with(
            mock.sentinel.glob_name)

    @mock.patch.object(coordination.COORDINATOR, 'get_lock')
    def test_synchronized_multiple_templates(self, get_lock):
        """Test locks requested in the right order and duplicates removed."""
        locks = ['lock-{f_name}-%s-{foo.val}-{bar[val]}' % i for i in range(3)]
        expect = [f'lock-func-{i}-7-8' for i in range(3)]

        @coordination.synchronized(locks[1], locks[0], locks[1], locks[2])
        def func(foo, bar):
            pass

        foo = mock.Mock(val=7)
        bar = mock.MagicMock()
        bar.__getitem__.return_value = 8
        func(foo, bar)
        self.assertEqual(len(expect), get_lock.call_count)
        get_lock.assert_has_calls([mock.call(lock) for lock in expect])

        self.assertEqual(len(expect), get_lock.return_value.acquire.call_count)
        get_lock.return_value.acquire.assert_has_calls(
            [mock.call(True)] * len(expect))

        self.assertEqual(['foo', 'bar'], inspect.getfullargspec(func)[0])

    @mock.patch('oslo_utils.timeutils.now', side_effect=[1, 2])
    def test___acquire(self, mock_now):
        lock = mock.Mock()
        # Using getattr to avoid AttributeError: module 'cinder.coordination'
        # has no attribute '_CoordinationTestCase__acquire'
        res = getattr(coordination, '__acquire')(lock, mock.sentinel.blocking,
                                                 mock.sentinel.f_name)
        self.assertEqual(2, res)
        self.assertEqual(2, mock_now.call_count)
        mock_now.assert_has_calls([mock.call(), mock.call()])
        lock.acquire.assert_called_once_with(mock.sentinel.blocking)

    @mock.patch('oslo_utils.timeutils.now')
    def test___acquire_propagates_exception(self, mock_now):
        lock = mock.Mock()
        lock.acquire.side_effect = ValueError
        # Using getattr to avoid AttributeError: module 'cinder.coordination'
        # has no attribute '_CoordinationTestCase__acquire'
        self.assertRaises(ValueError,
                          getattr(coordination, '__acquire'),
                          lock, mock.sentinel.blocking, mock.sentinel.f_name)
        mock_now.assert_called_once_with()
        lock.acquire.assert_called_once_with(mock.sentinel.blocking)

    @mock.patch('oslo_utils.timeutils.now', return_value=2)
    def test___release(self, mock_now):
        lock = mock.Mock()
        # Using getattr to avoid AttributeError: module 'cinder.coordination'
        # has no attribute '_CoordinationTestCase__release'
        getattr(coordination, '__release')(lock, 1, mock.sentinel.f_name)

        mock_now.assert_called_once_with()
        lock.release.assert_called_once_with()

    @mock.patch('oslo_utils.timeutils.now')
    def test___release_ignores_exception(self, mock_now):
        lock = mock.Mock()
        lock.release.side_effect = ValueError
        # Using getattr to avoid AttributeError: module 'cinder.coordination'
        # has no attribute '_CoordinationTestCase__release'
        getattr(coordination, '__release')(lock, 1, mock.sentinel.f_name)

        mock_now.assert_not_called()
        lock.release.assert_called_once_with()
